"""
Classes for incremental expanding panel cross-validators. 
"""

import numpy as np
import pandas as pd

from macrosynergy.learning.splitters.base_splitters import WalkForwardPanelSplit


class ExpandingIncrementPanelSplit(WalkForwardPanelSplit):
    """
    Walk-forward cross-validator over a panel.

    Provides train/test indices to split data into train/test sets. The dataset is split
    so that subsequent training sets are expanded by a fixed number of time periods to
    incorporate the latest available information. Each training set is followed by a test
    set of fixed length.

    Parameters
    ----------
    train_intervals : int
        The number of time periods by which the previous training set is expanded.
        Default is 21.
    test_size : int
        The number of time periods forward of each training set to use in the associated
        test set. Default is 21.
    min_cids : int
        The minimum number of cross-sections required for the first training set. Default
        is 4. Either start_date or (min_cids, min_periods) must be provided. If
        both are provided, start_date takes precedence.
    min_periods : int
        The minimum number of time periods required for the first training set. Default is
        500. Either start_date or (min_cids, min_periods) must be provided. If
        both are provided, start_date takes precedence.
    start_date : Optional[str]
        The targeted final date in the initial training set in ISO 8601 format.
        Default is None. Either start_date or (min_cids, min_periods) must be provided.
        If both are provided, start_date takes precedence.
    max_periods : Optional[int]
        The maximum number of time periods in each training set. If the maximum is
        exceeded, the earliest periods are cut off. This effectively creates rolling
        training sets. Default is None.

    Notes
    -----
    The first training set is determined by the specification of either `start_date` or
    by the parameters `min_cids` and `min_periods`. When `start_date` is provided, the
    initial training set comprises all available data before and including the `start_date`,
    unless `max_periods` is specified, in which case at most the last `max_periods` periods
    prior to the `start_date` are included.

    If `start_date` is not provided, the first training set is determined by the
    parameters `min_cids` and `min_periods`. This set comprises at least `min_periods`
    time periods for at least `min_cids` cross-sections.
    """

    def __init__(
        self,
        train_intervals=21,
        test_size=21,
        min_cids=4,
        min_periods=500,
        start_date=None,
        max_periods=None,
    ):
        # Checks
        super().__init__(
            min_cids=min_cids,
            min_periods=min_periods,
            start_date=start_date,
            max_periods=max_periods,
        )
        self._check_init_params(
            train_intervals=train_intervals,
            test_size=test_size,
        )

        # Attributes
        self.train_intervals = train_intervals
        self.test_size = test_size

    def split(self, X, y, groups=None):
        """
        Generate indices to split data into training and test sets.

        Parameters
        ----------
        X : pd.DataFrame
            Pandas dataframe of features, multi-indexed by (cross-section, date). The
            dates must be in datetime format. Otherwise the dataframe must be in wide
            format: each feature is a column.
        y : Union[pd.DataFrame, pd.Series]
            Pandas dataframe or series of a target variable, multi-indexed by
            (cross-section, date). The dates must be in datetime format. If a dataframe
            is provided, the target variable must be the sole column.
        groups : None
            Ignored. Exists for compatibility with scikit-learn.

        Yields
        ------
        train : np.ndarray
            The training set indices for that split.
        test : np.ndarray
            The testing set indices for that split.
        """
        # Checks
        self._check_split_params(X, y, groups)

        # Determine the unique dates in each training split
        train_indices = []
        test_indices = []

        Xy = pd.concat([X, y], axis=1).dropna()
        real_dates = Xy.index.get_level_values(1)
        splits = self._determine_unique_training_times(Xy, real_dates)

        # Determine the training and test indices for each split
        train_splits: list = [
            splits[0] if not self.max_periods else splits[0][-self.max_periods :]
        ]
        for i in range(1, self.n_splits):
            train_splits.append(np.concatenate([train_splits[i - 1], splits[i]]))

            # Drop beginning of training set if it exceeds max_periods.
            if self.max_periods:
                train_splits[i] = train_splits[i][-self.max_periods :]

        for split in train_splits:
            train_indices: list = np.where(real_dates.isin(split))[0]
            test_start: int = self.unique_dates.get_loc(split.max()) + 1
            test_indices: list = np.where(
                real_dates.isin(
                    self.unique_dates[test_start : test_start + self.test_size]
                )
            )[0]

            yield train_indices, test_indices

    def _determine_unique_training_times(self, Xy, real_dates):
        """
        Returns the unique dates in each training split.

        Parameters
        ----------
        Xy : pd.DataFrame
            Combined pandas dataframe of features and the target variable, multi-indexed
            by (cross-section, date).
        real_dates : pd.Index
            The dates associated with each sample in Xy.

        Notes
        -----
        This method is called by self.split(). It also returns other variables needed
        for ensuing components of the self.split() method.
        """
        self.unique_dates: pd.DatetimeIndex = real_dates.unique().sort_values()

        # First determine the dates for the first training set
        if self.start_date:
            date_last_train = self.start_date
        else:
            init_mask = Xy.groupby(level=1).size().sort_index() >= self.min_cids
            date_first_min_cids: pd.Timestamp = (
                init_mask[init_mask == True].reset_index().real_date.min()
            )
            date_last_train: pd.Timestamp = self.unique_dates[
                self.unique_dates.get_loc(date_first_min_cids) + self.min_periods - 1
            ]

        # Determine all remaining training splits
        splits = [self.unique_dates[self.unique_dates <= date_last_train]]
        i = self.unique_dates.get_loc(date_last_train)

        # Loop until no test sets can be created
        while i < len(self.unique_dates) - (self.test_size + 1):
            next_loc = i + self.train_intervals
            splits.append(self.unique_dates[i + 1 : next_loc + 1])
            i = next_loc

        self.n_splits = len(splits)

        return splits

    def get_n_splits(self, X, y, groups=None):
        """
        Calculates and returns the number of splits.

        Parameters
        ----------
        X : pd.DataFrame
            Pandas dataframe of features, multi-indexed by (cross-section, date). The
            dates must be in datetime format. Otherwise the dataframe must be in wide
            format: each feature is a column.

        y : Union[pd.DataFrame, pd.Series]
            Pandas dataframe or series of a target variable, multi-indexed by
            (cross-section, date). The dates must be in datetime format. If a dataframe
            is provided, the target variable must be the sole column.

        groups : None
            Ignored. Exists for compatibility with scikit-learn.

        Returns
        -------
        n_splits : int
            The number of splits.
        """
        Xy = pd.concat([X, y], axis=1)
        Xy.dropna(inplace=True)
        real_dates = Xy.index.get_level_values(1)
        self._determine_unique_training_times(Xy, real_dates)

        return self.n_splits

    def _check_init_params(
        self,
        train_intervals,
        test_size,
    ):
        """
        Type and value checks for the class initialisation parameters.

        Parameters
        ----------
        train_intervals : int
            The number of time periods by which the previous training set is expanded.
        test_size : int
            The number of time periods forward of each training set to use in the associated
            test set.
        """
        # train_intervals
        if not isinstance(train_intervals, int):
            raise TypeError(
                f"train_intervals must be an integer. Got {type(train_intervals)}."
            )
        if train_intervals < 1:
            raise ValueError(
                f"train_intervals must be an integer greater than 0. Got {train_intervals}."
            )

        # test_size
        if not isinstance(test_size, int):
            raise TypeError(f"test_size must be an integer. Got {type(test_size)}.")
        if test_size < 1:
            raise ValueError(
                f"test_size must be an integer greater than 0. Got {test_size}."
            )


class ExpandingFrequencyPanelSplit(WalkForwardPanelSplit):
    """
    Walk-forward cross-validator over a panel.

    Provides train/test indices to split data into train/test sets. The dataset is split
    so that subsequent training sets are expanded by a user-specified frequency to
    incorporate the latest available information. Each training set is followed by a test
    set spanning a user-defined frequency.

    Parameters
    ----------
    expansion_freq : str
        Frequency of training set expansion. For a given native dataset frequency, the
        training sets expand by the smallest number of dates to cover this frequency.
        Default is "D". Accepted values are "D", "W", "M", "Q" and "Y".
    test_freq : str
        Frequency forward of each training set for the unique dates in each test set to
        cover. Default is "D". Accepted values are "D", "W", "M", "Q" and "Y".
    min_cids : int
        Minimum number of cross-sections required for the initial training set.
        Default is 4. Either start_date or (min_cids, min_periods, min_xcats) must be
        provided. If both are provided, start_date takes precedence.
    min_periods : int
        Minimum number of time periods required for the initial training set. Default is
        500. Either start_date or (min_cids, min_periods, min_xcats) must be provided. If
        both are provided, start_date takes precedence.
    start_date : Optional[str]
        First rebalancing date in ISO 8601 format. This is the last date of the first
        training set. Default is None. Either start_date or (min_cids, min_periods)
        must be provided. If both are provided, start_date takes precedence.
    max_periods : Optional[int]
        The maximum number of time periods in each training set. If the maximum is
        exceeded, the earliest periods are cut off. This effectively creates rolling
        training sets. Default is None.

    Notes
    -----
    The first training set is either determined by the specification of `start_date` or by
    the parameters `min_cids` and `min_periods` collectively. When
    `start_date` is provided, the initial training set comprises all available data prior
    to the `start_date`, unless `max_periods` is specified, in which case at most the last
    `max_periods` periods prior to the `start_date` are included.

    If `start_date` is not provided, the first training set is determined by the parameters
    `min_cids` and `min_periods`. This set comprises at least `min_periods`
    time periods for at least `min_cids` cross-sections.

    This initial training set is immediately adjusted depending on the specified training
    interval frequency. For instance, if the training frequency is "M", the initial
    training set is further expanding so that all samples prior to the end of the month
    are included.

    The associated test set immediately follows the adjusted initial training set and spans
    the specified test set frequency forward of its associated training set.
    For instance, if the test frequency is "Q", the available dates that cover the
    subsequent quarter are grouped together to form the test set.

    Subsequent training sets are created by expanding the previous training set by the
    smallest number of dates to cover the training frequency. As before, each test set immediately
    follows its associated training set and is determined in the same manner as the initial test set.
    """

    def __init__(
        self,
        expansion_freq="D",
        test_freq="D",
        min_cids=4,
        min_periods=500,
        start_date=None,
        max_periods=None,
    ):
        # Checks
        super().__init__(
            min_cids=min_cids,
            min_periods=min_periods,
            start_date=start_date,
            max_periods=max_periods,
        )
        self._check_init_params(
            expansion_freq=expansion_freq,
            test_freq=test_freq,
        )

        # Attributes
        self.expansion_freq = expansion_freq
        self.test_freq = test_freq
        self.freq_offsets = {
            "D": pd.DateOffset(days=1),
            "W": pd.DateOffset(weeks=1),
            "M": pd.DateOffset(months=1),
            "Q": pd.DateOffset(months=3),
            "Y": pd.DateOffset(years=1),
        }

    def split(self, X, y, groups=None):
        """Generate indices to split data into training and test sets.

        Parameters
        ----------
        X : pd.DataFrame
            Pandas dataframe of features, multi-indexed by (cross-section, date). The
            dates must be in datetime format. Otherwise the dataframe must be in wide
            format: each feature is a column.

        y : Union[pd.DataFrame, pd.Series]
            Pandas dataframe or series of a target variable, multi-indexed by
            (cross-section, date). The dates must be in datetime format. If a dataframe
            is provided, the target variable must be the sole column.

        groups : None
            Ignored. Exists for compatibility with scikit-learn.

        Yields
        ------
        train : np.ndarray
            The training set indices for that split.

        test : np.ndarray
            The testing set indices for that split.
        """
        # Checks
        self._check_split_params(X, y, groups)

        # Determine the unique dates in each training split
        train_indices = []
        test_indices = []

        Xy = pd.concat([X, y], axis=1).dropna()
        real_dates = Xy.index.get_level_values(1)
        splits = self._determine_unique_training_times(Xy, real_dates)

        train_splits: list = [
            splits[0] if not self.max_periods else splits[0][-self.max_periods :]
        ]
        for i in range(1, self.n_splits):
            train_splits.append(np.concatenate([train_splits[i - 1], splits[i]]))

            # Drop beginning of training set if it exceeds max_periods.
            if self.max_periods:
                train_splits[i] = train_splits[i][-self.max_periods :]

        test_offset = self.freq_offsets[self.test_freq]
        for split_idx, split in enumerate(train_splits):
            train_indices = np.where(Xy.index.get_level_values(1).isin(split))[0]
            test_start: pd.Timestamp = self.unique_dates[
                self.unique_dates.get_loc(split.max()) + 1
            ]
            test_end: pd.Timestamp = test_start + test_offset
            if split_idx == len(train_splits) - 1:
                test_dates = sorted(self.unique_dates[self.unique_dates >= test_start])
            else:
                test_dates = sorted(
                    self.unique_dates[
                        (self.unique_dates >= test_start)
                        & (self.unique_dates < test_end)
                    ]
                )
            test_indices = np.where(Xy.index.get_level_values(1).isin(test_dates))[0]

            yield train_indices, test_indices

    def get_n_splits(self, X=None, y=None, groups=None) -> int:
        """
        Calculates and returns the number of splits.

        :param <pd.DataFrame> X: Pandas dataframe of features,
            multi-indexed by (cross-section, date). The dates must be in datetime format.
            Otherwise the dataframe must be in wide format: each feature is a column.
        :param <pd.DataFrame> y: Pandas dataframe of the target variable, multi-indexed by
            (cross-section, date). The dates must be in datetime format.
        :param <pd.DataFrame> groups: Always ignored, exists for compatibility.

        :return <int> n_splits: Returns the number of splits.
        """
        self._determine_unique_time_splits(X, y)

        return self.n_splits

    def _determine_unique_training_times(self, Xy, real_dates):
        """
        Returns the unique dates in each training split.

        Parameters
        ----------
        Xy : pd.DataFrame
            Combined pandas dataframe of features and the target variable, multi-indexed
            by (cross-section, date).
        real_dates : pd.Index
            The dates associated with each sample in Xy.

        Notes
        -----
        This method is called by self.split(). It also returns other variables needed
        for ensuing components of the self.split() method.
        """
        self.unique_dates: pd.DatetimeIndex = real_dates.unique().sort_values()
        # First determine the dates for the first training set
        if self.start_date:
            date_last_train = self.start_date
        else:
            init_mask = Xy.groupby(level=1).size().sort_index() >= self.min_cids
            date_first_min_cids: pd.Timestamp = (
                init_mask[init_mask == True].reset_index().real_date.min()
            )
            date_last_train: pd.Timestamp = self.unique_dates[
                self.unique_dates.get_loc(date_first_min_cids) + self.min_periods - 1
            ]

        # Determine all remaining training splits
        # To do this, loop through unique panel dates to create the training and test sets
        # Loop until the last "test_freq" dates in the panel are reached
        end_date = self.unique_dates[-1] - self.freq_offsets[self.test_freq]
        unique_dates_train: pd.arrays.DatetimeArray = self.unique_dates[
            (self.unique_dates > date_last_train) & (self.unique_dates <= end_date)
        ].sort_values()
        current_date = unique_dates_train[0]
        splits = []

        while current_date < end_date:
            next_date = current_date + self.freq_offsets[self.expansion_freq]
            if next_date > end_date:
                mask = (unique_dates_train >= current_date) & (
                    unique_dates_train <= end_date
                )
            else:
                mask = (unique_dates_train >= current_date) & (
                    unique_dates_train < next_date
                )
            split_dates = unique_dates_train[mask]
            if not split_dates.empty:
                splits.append(split_dates)
            current_date = next_date

        # Add the first training set to the list of training splits, so that the dates
        # that constitute each training split are together.
        splits.insert(
            0,
            real_dates[real_dates <= date_last_train].unique().sort_values(),
        )

        self.n_splits = len(splits)

        return splits

    def _check_init_params(
        self,
        expansion_freq,
        test_freq,
    ):
        # expansion_freq
        if not isinstance(expansion_freq, str):
            raise TypeError(
                f"expansion_freq must be a string. Got {type(expansion_freq)}."
            )
        if expansion_freq not in ["D", "W", "M", "Q", "Y"]:
            raise ValueError(
                f"expansion_freq must be one of 'D', 'W', 'M', 'Q' or 'Y'."
                f" Got {expansion_freq}."
            )

        # test_freq
        if not isinstance(test_freq, str):
            raise TypeError(f"test_freq must be a string. Got {type(test_freq)}.")
        if test_freq not in ["D", "W", "M", "Q", "Y"]:
            raise ValueError(
                f"test_freq must be one of 'D', 'W', 'M', 'Q' or 'Y'."
                f" Got {test_freq}."
            )


if __name__ == "__main__":
    from macrosynergy.management.simulate import make_qdf
    import macrosynergy.management as msm

    # Example dataset
    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2001-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {"GBP": ["2009-01-01", "2012-06-30"], "CAD": ["2018-01-01", "2100-01-01"]}
    dfd = msm.reduce_df(df=dfd, cids=cids, xcats=xcats, blacklist=black)

    dfd = dfd.pivot(index=["cid", "real_date"], columns="xcat", values="value")
    X = dfd.drop(columns=["XR"])
    y = dfd["XR"]

    # ExpandingIncrementPanelSplit
    splitter = ExpandingIncrementPanelSplit(train_intervals=21 * 12, test_size=21 * 12)
    splitter.visualise_splits(X, y)

    # ExpandingFrequencyPanelSplit
    splitter = ExpandingFrequencyPanelSplit(expansion_freq="Y", test_freq="Y")
    splitter.visualise_splits(X, y)



from .linear_model import (
    LADRegressor,
    SignWeightedLADRegressor,
    TimeWeightedLADRegressor,
    SignWeightedLinearRegression,
    TimeWeightedLinearRegression,
    ModifiedLinearRegression,
    ModifiedSignWeightedLinearRegression,
    ModifiedTimeWeightedLinearRegression,
)

from .model_systems import (
    CorrelationVolatilitySystem,
    LADRegressionSystem,
    LinearRegressionSystem,
    RidgeRegressionSystem,
)

from .naive_predictors import (
    NaiveRegressor,
)

from .neighbors.nearest_neighbors import KNNClassifier

from .meta_estimators import ProbabilityEstimator, FIExtractor

from .ensemble import (
    VotingClassifier,
    VotingRegressor,
)

__all__ = [
    "LADRegressor",
    "KNNClassifier",
    "SignWeightedLADRegressor",
    "TimeWeightedLADRegressor",
    "SignWeightedLinearRegression",
    "TimeWeightedLinearRegression",
    "NaiveRegressor",
    "CorrelationVolatilitySystem",
    "LADRegressionSystem",
    "LinearRegressionSystem",
    "RidgeRegressionSystem",
    "ModifiedLinearRegression",
    "ModifiedSignWeightedLinearRegression",
    "ModifiedTimeWeightedLinearRegression",
    "ProbabilityEstimator",
    "VotingClassifier",
    "VotingRegressor",
    "FIExtractor",
]


from .base_panel_scaler import (
    BasePanelScaler,
)

from .scalers import (
    PanelMinMaxScaler,
    PanelStandardScaler,
)

__all__ = [
    "BasePanelScaler",
    "PanelMinMaxScaler",
    "PanelStandardScaler",
]


"""
Base classes for panel cross-validation splitters.
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import datetime

from sklearn.model_selection import (
    BaseCrossValidator,
)

from abc import ABC, abstractmethod


class BasePanelSplit(BaseCrossValidator, ABC):
    """
    Generic cross-validation class for panel data.

    Notes
    -----
    This class is designed to provide a common interface for cross-validation on panel
    data. Much of the logic can be written in subclasses, but this base class contains
    the necessary code to visualise the splits for each cross-section in the panel.
    """

    @abstractmethod
    def __init__(self):
        """
        Constructor for the base class.
        """
        pass

    def get_n_splits(self, X=None, y=None, groups=None):
        """
        Returns the number of splits in the cross-validator.

        Parameters
        ----------
        X : pd.DataFrame, optional
            Pandas dataframe of features, multi-indexed by (cross-section, date). The
            dates must be in datetime format. Otherwise the dataframe must be in wide
            format: each feature is a column.
        y : Union[pd.Series, pd.DataFrame], optional
            Pandas dataframe or series of a target variable, multi-indexed by
            (cross-section, date). The dates must be in datetime format. If a dataframe
            is provided, the target variable must be the sole column.
        groups : None
            Always ignored, exists for compatibility with scikit-learn.

        Returns
        -------
        n_splits : int
            Number of splits in the cross-validator.
        """
        # Checks
        self._check_Xy(X, y)

        # Determine number of splits in the cross-validator
        if hasattr(self, "n_splits"):
            return self.n_splits
        else:
            return len(list(self.split(X, y)))

    def _calculate_xranges(
        self,
        cs_dates,
        real_dates,
        freq_offset,
    ):
        """
        Returns date ranges of contiguous blocks in each training and test set.

        Parameters
        ----------
        cs_dates : pd.DatetimeIndex
            DatetimeIndex of dates in a (training or test) set for a given cross-section.
        real_dates : pd.DatetimeIndex
            DatetimeIndex of all dates in the panel.
        freq_offset : pd.DateOffset
            DateOffset object representing the frequency of the dates in the panel.

        Returns
        -------
        xranges : tuple
            List of tuples of the form (start date, length of contiguous dates).
        """
        xranges = []
        if len(cs_dates) == 0:
            # No dates in the training or test set - return empty list.
            return xranges

        # Filter all dates in the panel spanning the dates in the considered set
        filtered_real_dates = real_dates[
            (real_dates >= cs_dates.min()) & (real_dates <= cs_dates.max())
        ]
        # Differences may arise due to blacklisting
        difference = filtered_real_dates.difference(cs_dates)

        if len(difference) == 0:
            # Only one contiguous range of dates.
            xranges.append(
                (cs_dates.min(), cs_dates.max() + freq_offset - cs_dates.min())
            )
        else:
            # Multiple contiguous blocks present. Iterate over them.
            while len(difference) > 0:
                xranges.append((cs_dates.min(), difference.min() - cs_dates.min()))
                cs_dates = cs_dates[(cs_dates >= difference.min())]
                difference = difference[(difference >= cs_dates.min())]

            xranges.append(
                (cs_dates.min(), cs_dates.max() + freq_offset - cs_dates.min())
            )

        return xranges

    def visualise_splits(
        self,
        X,
        y,
        figsize=(20, 5),
        show_title=True,
        tick_fontsize=None,
        label_fontsize=None,
        subtitle_fontsize=None,
    ):
        """
        Visualise the cross-validation splits.

        Parameters
        ----------
        X : pd.DataFrame
            Pandas dataframe of features/quantamental indicators, multi-indexed by
            (cross-section, date). The dates must be in datetime format. The
            dataframe must be in wide format: each feature is a column.
        y : pd.DataFrame
            Pandas dataframe of target variable, multi-indexed by (cross-section, date).
            The dates must be in datetime format.
        figsize : Tuple[int, int]
            Tuple of integers specifying the splitter visualisation figure size.
        show_title : bool, optional
            Boolean specifying whether to show the title of the figure. Default is True.
        tick_fontsize : int, optional
            Integer specifying the size of the x-axis tick labels. Default is None.
        label_fontsize : int, optional
            Integer specifying the size of the y-axis labels. Default is None.
        subtitle_fontsize : int, optional
            Integer specifying the size of the subplot titles. Default is None.
        """
        sns.set_theme(style="whitegrid", palette="colorblind")

        # Checks
        self._check_Xy(X, y)
        if not isinstance(figsize, tuple):
            raise TypeError("figsize must be a tuple.")
        for i in figsize:
            if not isinstance(i, int):
                raise TypeError("figsize must contain only integers.")
            if i <= 0:
                raise ValueError("figsize must contain only positive integers.")

        if tick_fontsize is not None:
            if not isinstance(tick_fontsize, int):
                raise TypeError("ticksize must be an integer.")

        if label_fontsize is not None:
            if not isinstance(label_fontsize, int):
                raise TypeError("labelsize must be an integer.")
        
        if subtitle_fontsize is not None:
            if not isinstance(subtitle_fontsize, int):
                raise TypeError("subtitle_size must be an integer.")

        # Obtain relevant data
        Xy: pd.DataFrame = pd.concat([X, y], axis=1).dropna()
        cross_sections = np.array(sorted(Xy.index.get_level_values(0).unique()))
        real_dates = Xy.index.get_level_values(1).unique().sort_values()

        # Infer native dataset frequency
        freq_est = pd.infer_freq(real_dates)
        if freq_est is None:
            freq_est = real_dates.to_series().diff().min()
        freq_offset = pd.tseries.frequencies.to_offset(freq_est)  # Good approximation

        splits = list(self.split(X, y))

        # Set up plotting labels and figure
        split_idxs: list = (
            [0, len(splits) // 4, len(splits) // 2, 3 * len(splits) // 4, -1]
            if self.n_splits > 5
            else [i for i in range(self.n_splits)]
        )
        split_titles: list = (
            [
                "Initial split",
                "Quarter progress",
                "Halfway progress",
                "Three quarters progress",
                "Final split",
            ]
            if self.n_splits > 5
            else [f"Split {i+1}" for i in range(self.n_splits)]
        )
        fig, ax = plt.subplots(
            nrows=len(cross_sections),
            ncols=min(self.n_splits, 5),
            figsize=figsize,
        )

        # Determine ranges of contiguous dates in each training and test set, within each
        # cross-section and split.
        plot_components = []
        for cs_idx, cs in enumerate(cross_sections):
            for idx, split_idx in enumerate(split_idxs):
                # Get the dates in the training and test sets for the given cross-section.
                cs_train_dates: pd.DatetimeIndex = Xy.iloc[splits[split_idx][0]][
                    Xy.iloc[splits[split_idx][0]].index.get_level_values(0) == cs
                ].index.get_level_values(1)
                cs_test_dates: pd.DatetimeIndex = Xy.iloc[splits[split_idx][1]][
                    Xy.iloc[splits[split_idx][1]].index.get_level_values(0) == cs
                ].index.get_level_values(1)

                xranges_train: list = self._calculate_xranges(
                    cs_train_dates, real_dates, freq_offset
                )
                xranges_test: list = self._calculate_xranges(
                    cs_test_dates, real_dates, freq_offset
                )

                plot_components.append(
                    (cs_idx, idx, xranges_train, "royalblue", "Train")
                )
                plot_components.append(
                    (cs_idx, idx, xranges_test, "lightcoral", "Test")
                )

        # Calculate the difference between final two dates.
        # This will be added to the x-axis limits to ensure that the final split is visible.
        last_date = real_dates[-1]
        second_to_last_date = real_dates[-2]
        difference = last_date - second_to_last_date

        # Add all the broken bar plots to the figure.
        for cs_idx, idx, xranges, color, label in plot_components:
            if len(cross_sections) == 1:
                ax[idx].broken_barh(xranges, (-0.4, 0.8), facecolors=color, label=label)
                ax[idx].set_xlim(real_dates.min(), real_dates.max() + difference)
                ax[idx].set_yticks([0])
                ax[idx].set_yticklabels([cross_sections[0]], fontsize=label_fontsize)
                ax[idx].tick_params(axis="x", rotation=90, labelsize=tick_fontsize)
                ax[idx].set_title(f"{split_titles[idx]}", fontsize=subtitle_fontsize)
            elif len(split_idxs) == 1:
                ax[cs_idx].broken_barh(
                    xranges, (-0.4, 0.8), facecolors=color, label=label
                )
                ax[cs_idx].set_xlim(real_dates.min(), real_dates.max() + difference)
                ax[cs_idx].set_yticks([0])
                ax[cs_idx].set_yticklabels([cross_sections[cs_idx]], fontsize=label_fontsize)
                ax[cs_idx].tick_params(axis="x", rotation=90, labelsize=tick_fontsize)
            else:
                ax[cs_idx, idx].broken_barh(
                    xranges, (-0.4, 0.8), facecolors=color, label=label
                )
                ax[cs_idx, idx].set_xlim(
                    real_dates.min(), real_dates.max() + difference
                )
                ax[cs_idx, idx].set_yticks([0])
                ax[cs_idx, idx].set_yticklabels(
                    [cross_sections[cs_idx]], fontsize=label_fontsize
                )
                ax[cs_idx, idx].tick_params(axis="x", rotation=90, labelsize=tick_fontsize)

                # Ensure only the last row has x-axis labels.
                if cs_idx == len(ax) - 1:
                    ax[cs_idx, idx].tick_params(
                        axis="x", rotation=90, labelsize=tick_fontsize
                    )
                else:
                    ax[cs_idx, idx].tick_params(axis="x", labelbottom=False)

                if cs_idx == 0:
                    ax[cs_idx, idx].set_title(
                        f"{split_titles[idx]}", fontsize=subtitle_fontsize
                    )

        if show_title:
            plt.suptitle(
                f"Training and test set pairs, number of training sets={self.n_splits}"
            )
        plt.legend(frameon=True)
        plt.tight_layout()
        plt.show()

    def _check_Xy(self, X, y):
        """
        Type and value checks of input feature and target panels.

        Parameters
        ----------
        X : pd.DataFrame, optional
            Pandas dataframe of features/quantamental indicators, multi-indexed by
            (cross-section, date). The dates must be in datetime format. The
            dataframe must be in wide format: each feature is a column.
        y : pd.DataFrame, optional
            Pandas dataframe of target variable, multi-indexed by (cross-section, date).
            The dates must be in datetime format.
        """
        if X is None:
            # Check y is not provided
            if y is not None:
                raise ValueError("Either both X and y must be provided or neither.")
        else:
            # Check y is provided
            if y is None:
                raise ValueError("Either both X and y must be provided or neither.")

        if X is not None:
            # Check X and y are dataframes/series respectively
            if not isinstance(X, pd.DataFrame):
                raise TypeError("X must be a pandas dataframe.")
            if not isinstance(y, (pd.DataFrame, pd.Series)):
                raise TypeError("y must be a pandas dataframe or series.")
            if isinstance(y, pd.DataFrame) and len(y.columns) != 1:
                raise ValueError("If y is a dataframe, it must have only one column.")

            # Check indexing of X and y
            if not X.index.equals(y.index):
                raise ValueError("X and y must have the same index.")
            if not isinstance(X.index, pd.MultiIndex):
                raise ValueError("X and y must have a multi-index.")
            if X.index.get_level_values(0).dtype != "object":
                raise ValueError(
                    "The input data must have string outer index representing panel cross-sections."
                )
            if X.index.get_level_values(1).dtype != "datetime64[ns]":
                raise ValueError(
                    "The input data must have datetime inner index representing timestamps as datetime64[ns]."
                )


class WalkForwardPanelSplit(BasePanelSplit, ABC):
    """
    Generic walk-forward panel cross-validator.

    Parameters
    ----------
    min_cids : int
        Minimum number of cross-sections required for the first training set.
        Either start_date or (min_cids, min_periods) must be provided.
        If both are provided, start_date takes precedence.
    min_periods : int
        Minimum number of time periods required for the first training set. Either
        start_date or (min_cids, min_periods) must be provided. If both are
        provided, start_date takes precedence.
    start_date : str, optional
        The targeted final date in the initial training set in ISO 8601 format.
        Default is None. Either start_date or (min_cids, min_periods) must be provided.
        If both are provided, start_date takes precedence.
    max_periods : int, optional
        The maximum number of time periods in each training set. If the maximum is
        exceeded, the earliest periods are cut off. This effectively creates rolling
        training sets. Default is None.

    Notes
    -----
    Provides train/test indices to split a panel into train/test sets. Following an
    initial training set construction, a forward test set is created. The training and
    test set pair evolves over time by walking forward through the panel.
    """

    def __init__(
        self,
        min_cids,
        min_periods,
        start_date=None,
        max_periods=None,
    ):
        # Checks
        self._check_wf_params(
            min_cids=min_cids,
            min_periods=min_periods,
            start_date=start_date,
            max_periods=max_periods,
        )

        # Attributes
        self.min_cids = min_cids
        self.min_periods = min_periods
        self.start_date = pd.Timestamp(start_date) if start_date else None
        self.max_periods = max_periods

    def _check_wf_params(self, min_cids, min_periods, start_date, max_periods):
        """
        Type and value checks for the class initialisation parameters.

        Parameters
        ----------
        min_cids : int
            Minimum number of cross-sections required for the first training set.
        min_periods : int
            Minimum number of time periods required for the first training set.
        start_date : str
            The targeted final date in the initial training set in ISO 8601 format.
        max_periods : int
            The maximum number of time periods in each training set.
        """
        # min_cids
        if not isinstance(min_cids, int):
            raise TypeError(f"min_cids must be an integer. Got {type(min_cids)}.")
        if min_cids < 1:
            raise ValueError(
                f"min_cids must be an integer greater than 0. Got {min_cids}."
            )
        # min_periods
        if not isinstance(min_periods, int):
            raise TypeError(f"min_periods must be an integer. Got {type(min_periods)}.")
        if min_periods < 1:
            raise ValueError(
                f"min_periods must be an integer greater than 0. Got {min_periods}."
            )
        # start_date
        if start_date is not None and not isinstance(start_date, str):
            raise TypeError(f"start_date must be a string. Got {type(start_date)}.")
        if start_date is not None:
            try:
                datetime.datetime.fromisoformat(start_date)
            except ValueError:
                raise ValueError(
                    f"start_date must be in ISO 8601 format. Got {start_date}."
                )
        # max_periods
        if max_periods is not None and not isinstance(max_periods, int):
            raise TypeError(f"max_periods must be an integer. Got {type(max_periods)}.")
        if max_periods is not None and max_periods < 1:
            raise ValueError(
                f"max_periods must be an integer greater than 0. Got {max_periods}."
            )

    def _check_split_params(self, X, y, groups):
        """
        Type and value checks for the `split()` method parameters.

        Parameters
        ----------
        X : pd.DataFrame
            Pandas dataframe of features, multi-indexed by (cross-section, date). The
            dates must be in datetime format. Otherwise the dataframe must be in wide
            format: each feature is a column.

        y : pd.DataFrame or pd.Series
            Pandas dataframe or series of a target variable, multi-indexed by
            (cross-section, date). The dates must be in datetime format. If a dataframe
            is provided, the target variable must be the sole column.

        groups : None
            Ignored. Exists for compatibility with scikit-learn.
        """
        # X
        if not isinstance(X, pd.DataFrame):
            raise TypeError("X must be a pandas dataframe.")
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not pd.api.types.is_datetime64_any_dtype(X.index.get_level_values(1)):
            raise ValueError(
                f"The dates in X must be datetime objects. Got {X.index.get_level_values(1).dtype} instead."
            )
        # y
        if not isinstance(y, (pd.DataFrame, pd.Series)):
            raise TypeError("y must be a pandas dataframe or series.")
        if not isinstance(y.index, pd.MultiIndex):
            raise ValueError("y must be multi-indexed.")
        if not pd.api.types.is_datetime64_any_dtype(y.index.get_level_values(1)):
            raise ValueError(
                f"The dates in y must be datetime objects. Got {y.index.get_level_values(1).dtype} instead."
            )
        if not X.index.equals(y.index):
            raise ValueError(
                "The indices of the input dataframe X and the output dataframe y don't"
                "match."
            )
        # groups
        if groups is not None:
            raise ValueError("groups is not supported by this splitter.")


class KFoldPanelSplit(BasePanelSplit, ABC):
    def __init__(self, n_splits=5, min_n_splits=2):
        """
        Generic K-Fold cross-validator for panel data.

        Parameters
        ----------
        n_splits : int
            Number of splits to generate.
        min_n_splits : int
            Minimum number of splits allowed.

        Notes
        -----
        Provides train/test indices to split a panel into train/test sets. The panel is
        divided into n_splits consecutive folds. Each fold is then used once as a
        validation fold whilst a proportion of the other folds are used as training data.
        """
        # Checks
        if not isinstance(n_splits, int):
            raise TypeError(f"n_splits must be an integer. Got {type(n_splits)}.")
        if n_splits < min_n_splits:
            raise ValueError(
                f"Cannot have number of splits less than {min_n_splits}. Got n_splits = {n_splits}."
            )

        # Attributes
        self.n_splits = n_splits

    def split(self, X, y, groups=None):
        """
        Generate indices to split data into training and test sets.

        Parameters
        ----------
        X : pd.DataFrame
            Pandas dataframe of features, multi-indexed by (cross-section, date). The
            dates must be in datetime format. Otherwise the dataframe must be in wide
            format: each feature is a column.

        y : Union[pd.DataFrame, pd.Series]
            Pandas dataframe or series of a target variable, multi-indexed by
            (cross-section, date). The dates must be in datetime format. If a dataframe
            is provided, the target variable must be the sole column.

        groups : None
            Ignored. Exists for compatibility with scikit-learn.

        Yields
        ------
        train : np.ndarray
            The training set indices for that split.

        test : np.ndarray
            The testing set indices for that split.
        """
        # Checks
        self._check_split_params(X, y, groups)

        # Store necessary quantities
        Xy = pd.concat([X, y], axis=1)
        Xy.dropna(inplace=True)
        dates = Xy.index.get_level_values(1)
        unique_dates = dates.unique().sort_values()

        # Calculate splits
        splits = self._determine_splits(unique_dates, self.n_splits)

        # Yield splits
        for n_split in range(self.n_splits):
            yield self._get_split_indicies(n_split, splits, Xy, dates, unique_dates)

    @abstractmethod
    def _determine_splits(self, unique_dates, n_splits):
        """
        Determine panel splits based on the sorted collection of unique dates and the
        number of splits specified by the user.

        Parameters
        ----------
        unique_dates : pd.DatetimeIndex
            Sorted collection of unique dates in the panel.
        n_splits : int
            Number of splits to generate.

        Returns
        -------
        splits : list of np.ndarray
            List of numpy arrays denoting dates in each split.
        """
        pass

    @abstractmethod
    def _get_split_indicies(self, idx, split, splits, Xy, dates, unique_dates):
        """
        Determine the training and test set indices for a given split.

        Parameters
        ----------
        n_split : int
            Index of the current split.
        splits : list of np.ndarray
            List of numpy arrays denoting dates in each split.
        Xy : pd.DataFrame
            Combined dataframe of the features and the target variable.
        dates : pd.DatetimeIndex
            DatetimeIndex of all dates in the panel.
        unique_dates : pd.DatetimeIndex
            Sorted collection of unique dates in the panel.

        Returns
        -------
        train : np.ndarray
            The training set indices for that split.
        test : np.ndarray
            The testing set indices for that split.
        """
        pass

    def _check_split_params(self, X, y, groups):
        """
        Splitter input checks.

        Parameters
        ----------
        X : pd.DataFrame
            Pandas dataframe of features, multi-indexed by (cross-section, date). The
            dates must be in datetime format. Otherwise the dataframe must be in wide
            format: each feature is a column.

        y : Union[pd.DataFrame, pd.Series]
            Pandas dataframe or series of a target variable, multi-indexed by
            (cross-section, date). The dates must be in datetime format. If a dataframe
            is provided, the target variable must be the sole column.

        groups : None
            Ignored. Exists for compatibility with scikit-learn.
        """
        # X
        if not isinstance(X, pd.DataFrame):
            raise TypeError("X must be a pandas dataframe.")
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not pd.api.types.is_datetime64_any_dtype(X.index.get_level_values(1)):
            raise ValueError(
                f"The dates in X must be datetime objects. Got {X.index.get_level_values(1).dtype} instead."
            )
        # y
        if not isinstance(y, (pd.DataFrame, pd.Series)):
            raise TypeError("y must be a pandas dataframe or series.")
        if not isinstance(y.index, pd.MultiIndex):
            raise ValueError("y must be multi-indexed.")
        if not pd.api.types.is_datetime64_any_dtype(y.index.get_level_values(1)):
            raise ValueError(
                f"The dates in y must be datetime objects. Got {y.index.get_level_values(1).dtype} instead."
            )
        if not X.index.equals(y.index):
            raise ValueError(
                "The indices of the input dataframe X and the output dataframe y don't"
                "match."
            )
        # groups
        if groups is not None:
            raise ValueError("groups is not supported by this splitter.")



"""
Module hosting custom types and meta-classes for use across the package.
"""

from typing import Optional, Mapping, Union, Sequence, List, Tuple, Dict
import pandas as pd
from macrosynergy.compat import PD_2_0_OR_LATER
from .methods import (
    get_col_sort_order,
    change_column_format,
    _get_tickers_series,
    reduce_df,
    reduce_df_by_ticker,
    update_df,
    apply_blacklist,
    qdf_to_wide_df,
    check_is_categorical,
    add_nan_series,
    drop_nan_series,
    rename_xcats,
    qdf_from_timeseries,
    create_empty_categorical_qdf,
    concat_qdfs,
)
from .base import QuantamentalDataFrameBase


class QuantamentalDataFrame(QuantamentalDataFrameBase):
    """
    Type extension of `pd.DataFrame` for Quantamental DataFrames.  Usage: >>> df:
    pd.DataFrame = load_data() >>> qdf = QuantamentalDataFrame(df)
    """

    def __init__(
        self,
        df: Optional[pd.DataFrame] = None,
        categorical: bool = True,
        _initialized_as_categorical: Optional[bool] = None,
    ):
        if df is not None:
            df_err = "Input must be a standardised Quantamental DataFrame."
            if not isinstance(df, pd.DataFrame):
                raise TypeError(df_err)

            if not isinstance(df, QuantamentalDataFrame):
                if "real_date" in df.columns:
                    df["real_date"] = pd.to_datetime(df["real_date"])

                if not isinstance(df, QuantamentalDataFrame):
                    raise ValueError(df_err)

        if type(df) is QuantamentalDataFrame:
            if _initialized_as_categorical is None:
                _initialized_as_categorical = df.InitializedAsCategorical
        else:
            if df.columns.tolist() != get_col_sort_order(df):
                df = df[get_col_sort_order(df)]

        if PD_2_0_OR_LATER:
            super().__init__(df)
        else:
            super().__init__(df.copy()) # pragma: no cover

        _check_cat = check_is_categorical(self)
        if _initialized_as_categorical is None:
            self.InitializedAsCategorical = _check_cat
        else:
            if not isinstance(_initialized_as_categorical, bool):
                raise TypeError("`_initialized_as_categorical` must be a boolean.")
            self.InitializedAsCategorical = _initialized_as_categorical

        if categorical:
            if not _check_cat:
                self.to_categorical()
        else:
            self.to_string_type()

    def is_categorical(self) -> bool:
        """
        Returns True if the QuantamentalDataFrame is categorical.

        Returns
        -------
        bool
            True if the QuantamentalDataFrame is categorical
        """
        return check_is_categorical(self)

    def to_categorical(self) -> "QuantamentalDataFrame":
        """
        Converts the QuantamentalDataFrame to a categorical DataFrame.

        Returns
        -------
        QuantamentalDataFrame
            The QuantamentalDataFrame with categorical columns.
        """
        return change_column_format(self, cols=self._StrIndexCols, dtype="category")

    def to_string_type(self) -> "QuantamentalDataFrame":
        """
        Converts the QuantamentalDataFrame to a string DataFrame.

        Returns
        -------
        QuantamentalDataFrame
            The QuantamentalDataFrame with string columns.

        """
        return change_column_format(self, cols=self._StrIndexCols, dtype="object")

    def to_original_dtypes(self) -> "QuantamentalDataFrame":
        """
        Converts the QuantamentalDataFrame to its original dtypes (using the
        `InitialisedAsCategorical` attribute).

        Returns
        -------
        QuantamentalDataFrame
            The QuantamentalDataFrame with its original dtypes. The dtype is determined
            by the `InitialisedAsCategorical` attribute. The output dtype will be either
            'category' or 'object'.
        """

        if self.InitializedAsCategorical:
            return self.to_categorical()
        return self.to_string_type()

    def list_tickers(self) -> List[str]:
        """
        List all tickers in the QuantamentalDataFrame.

        Returns
        -------
        List[str]
            A list of all tickers in the QuantamentalDataFrame.
        """
        ltickers: List[str] = sorted(_get_tickers_series(self).unique())
        return ltickers

    def add_ticker_column(self) -> "QuantamentalDataFrame":
        """
        Add a ticker column to the QuantamentalDataFrame. `ticker` is a combination of
        `cid` and `xcat` columns. i.e. `ticker = cid_xcat`.

        Returns
        -------
        QuantamentalDataFrame
            The QuantamentalDataFrame with a `ticker` column.
        """
        ticker_col = _get_tickers_series(self)
        self["ticker"] = ticker_col
        return self

    def drop_ticker_column(self) -> "QuantamentalDataFrame":
        """
        Drop the ticker column from the QuantamentalDataFrame.

        Raises
        ------
        ValueError
            If no `ticker` column is found in the DataFrame.

        Returns
        -------
        QuantamentalDataFrame
            The QuantamentalDataFrame without the `ticker` column.
        """
        if "ticker" not in self.columns:
            raise ValueError("No `ticker` column found in the DataFrame.")
        return self.drop(columns=["ticker"])

    def reduce_df(
        self,
        cids: Optional[Sequence[str]] = None,
        xcats: Optional[Sequence[str]] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        blacklist: Dict[str, Sequence[Union[str, pd.Timestamp]]] = None,
        out_all: bool = False,
        intersect: bool = False,
    ) -> Union[
        "QuantamentalDataFrame", Tuple["QuantamentalDataFrame", List[str], List[str]]
    ]:
        """
        Filter DataFrame by `cids`, `xcats`, and `start` & `end` dates.

        Parameters
        ----------
        cids : Optional[Sequence[str]], optional
            List of CIDs to filter by, by default None
        xcats : Optional[Sequence[str]], optional
            List of XCATs to filter by, by default None
        start : Optional[str], optional
            Start date to filter by, by default None
        end : Optional[str], optional
            End date to filter by, by default None
        blacklist : Dict[str, Sequence[Union[str, pd.Timestamp]]], optional
            Blacklist to apply to the DataFrame, by default None
        out_all : bool, optional
            If True, return the filtered DataFrame, the filtered XCATs, and the filtered
            CIDs, by default False

        Returns
        -------
        QuantamentalDataFrame
            The filtered QuantamentalDataFrame. (if out_all=False, default)

        Tuple[QuantamentalDataFrame, List[str], List[str]]
            The filtered QuantamentalDataFrame, the filtered XCATs, and the filtered CIDs.
            (if out_all=True)
        """
        result = reduce_df(
            df=self,
            cids=cids,
            xcats=xcats,
            start=start,
            end=end,
            blacklist=blacklist,
            out_all=out_all,
            intersect=intersect,
        )

        if out_all:
            result, _xcats, _cids = result

        result = QuantamentalDataFrame(
            result,
            categorical=self.is_categorical(),
            _initialized_as_categorical=self.InitializedAsCategorical,
        )

        if out_all:
            return result, _xcats, _cids

        return result

    def reduce_df_by_ticker(
        self,
        tickers: Sequence[str],
        start: Optional[str] = None,
        end: Optional[str] = None,
        blacklist: Mapping[str, Sequence[Union[str, pd.Timestamp]]] = None,
    ) -> "QuantamentalDataFrame":
        """
        Filter DataFrame by `ticker`, `start` & `end` dates.

        Parameters
        ----------
        tickers : Sequence[str]
            List of tickers to filter by
        start : Optional[str], optional
            Start date to filter by, by default None
        end : Optional[str], optional
            End date to filter by, by default None
        blacklist : Mapping[str, Sequence[Union[str, pd.Timestamp]]], optional
            Blacklist to apply to the DataFrame, by default None

        Returns
        -------
        QuantamentalDataFrame
            The filtered QuantamentalDataFrame.
        """
        result = reduce_df_by_ticker(
            df=self,
            tickers=tickers,
            start=start,
            end=end,
            blacklist=blacklist,
        )
        return QuantamentalDataFrame(
            result,
            categorical=self.is_categorical(),
            _initialized_as_categorical=self.InitializedAsCategorical,
        )

    def apply_blacklist(
        self,
        blacklist: Mapping[str, Sequence[Union[str, pd.Timestamp]]],
    ):
        """
        Apply a blacklist to the QuantamentalDataFrame.
        """
        result = apply_blacklist(df=self, blacklist=blacklist)
        return QuantamentalDataFrame(
            result,
            categorical=self.is_categorical(),
            _initialized_as_categorical=self.InitializedAsCategorical,
        )

    def update_df(
        self,
        df_add: pd.DataFrame,
        xcat_replace: bool = False,
    ) -> "QuantamentalDataFrame":
        """
        Update the QuantamentalDataFrame with a new DataFrame.

        Parameters
        ----------
        df_add : pd.DataFrame
            DataFrame to update the QuantamentalDataFrame with
        xcat_replace : bool, optional
            If True, replace the XCATs in the QuantamentalDataFrame with the XCATs in
            `df_add`, by default False

        Returns
        -------
        QuantamentalDataFrame
            The updated QuantamentalDataFrame.
        """
        result = update_df(df=self, df_add=df_add, xcat_replace=xcat_replace)
        return QuantamentalDataFrame(
            result,
            categorical=self.is_categorical(),
            _initialized_as_categorical=self.InitializedAsCategorical,
        )

    def add_nan_series(
        self,
        ticker: str,
        start: Optional[str] = None,
        end: Optional[str] = None,
    ) -> "QuantamentalDataFrame":
        """
        Add a NaN series to the QuantamentalDataFrame.

        Parameters
        ----------
        ticker : str
            Ticker to add the NaN series to
        start : Optional[str], optional
            Start date of the NaN series, by default None
        end : Optional[str], optional
            End date of the NaN series, by default None

        Returns
        -------
        QuantamentalDataFrame
            The QuantamentalDataFrame with the NaN series added.
        """
        result = add_nan_series(df=self, ticker=ticker, start=start, end=end)
        return QuantamentalDataFrame(
            result,
            categorical=self.is_categorical(),
            _initialized_as_categorical=self.InitializedAsCategorical,
        )

    def drop_nan_series(
        self,
        column: str = "value",
        raise_warning: bool = True,
    ) -> "QuantamentalDataFrame":
        """
        Drop NaN series from the QuantamentalDataFrame.

        Parameters
        ----------
        column : str, optional
            Column to check for NaN series, by default "value"
        raise_warning : bool, optional
            If True, raise a warning if NaN series are dropped, by default True

        Returns
        -------
        QuantamentalDataFrame
            The QuantamentalDataFrame with NaN series dropped
        """
        result = drop_nan_series(df=self, column=column, raise_warning=raise_warning)
        return QuantamentalDataFrame(
            result,
            categorical=self.is_categorical(),
            _initialized_as_categorical=self.InitializedAsCategorical,
        )

    def rename_xcats(
        self,
        xcat_map: Optional[Mapping[str, str]] = None,
        select_xcats: Optional[List[str]] = None,
        postfix: Optional[str] = None,
        prefix: Optional[str] = None,
        name_all: Optional[str] = None,
        fmt_string: Optional[str] = None,
    ) -> "QuantamentalDataFrame":
        """
        Rename xcats in the QuantamentalDataFrame.

        Parameters
        ----------
        xcat_map : Optional[Mapping[str, str]], optional
            Mapping of xcats to rename, by default None
        select_xcats : Optional[List[str]], optional
            List of xcats to rename, by default None
        postfix : Optional[str], optional
            Postfix to add to the xcats, by default None
        prefix : Optional[str], optional
            Prefix to add to the xcats, by default None
        name_all : Optional[str], optional
            Name to rename all xcats to, by default None
        fmt_string : Optional[str], optional
            Format string to rename xcats, by default None

        Returns
        -------
        QuantamentalDataFrame
            The QuantamentalDataFrame with the xcats renamed.
        """
        result = rename_xcats(
            df=self,
            xcat_map=xcat_map,
            select_xcats=select_xcats,
            postfix=postfix,
            prefix=prefix,
            name_all=name_all,
            fmt_string=fmt_string,
        )
        return QuantamentalDataFrame(
            result,
            categorical=self.is_categorical(),
            _initialized_as_categorical=self.InitializedAsCategorical,
        )

    def to_wide(
        self,
        value_column: str = "value",
    ) -> "QuantamentalDataFrame":
        """
        Pivot the QuantamentalDataFrame.

        Parameters
        ----------
        value_column : str, optional
            Column to pivot, by default "value"

        Returns
        -------
        QuantamentalDataFrame
            The pivoted QuantamentalDataFrame, with each ticker as a column with the
            values of the `value_column` and the index as the `real_date`.
        """
        result = qdf_to_wide_df(self, value_column=value_column)
        if self.InitializedAsCategorical:
            return result

        result.columns = result.columns.astype("object")
        return result

    @classmethod
    def from_timeseries(
        cls,
        timeseries: pd.Series,
        ticker: str,
        metric: str = "value",
    ) -> "QuantamentalDataFrame":
        """
        Convert a timeseries DataFrame to a QuantamentalDataFrame.

        Parameters
        ----------
        timeseries : pd.Series
            Timeseries to convert to a QuantamentalDataFrame
        ticker : str
            Ticker to assign to the timeseries
        metric : str, optional
            Metric to assign to the timeseries, by default "value"

        Returns
        -------
        QuantamentalDataFrame
            The QuantamentalDataFrame created from the timeseries.
        """
        return QuantamentalDataFrame(
            qdf_from_timeseries(
                timeseries=timeseries,
                ticker=ticker,
                metric=metric,
            )
        )

    @classmethod
    def from_long_df(
        cls,
        df: pd.DataFrame,
        real_date_column: str = "real_date",
        value_column: str = "value",
        cid: Optional[str] = None,
        xcat: Optional[str] = None,
        ticker: Optional[str] = None,
        categorical: bool = True,
    ) -> "QuantamentalDataFrame":
        """
        Convert a long DataFrame to a QuantamentalDataFrame. This is useful when the
        DataFrame may contain only a `cid` or `xcat` column, or in cases where the `cid`
        and `xcat` columns are not named as such.

        Parameters
        ----------
        df : pd.DataFrame
            Long DataFrame to convert to a QuantamentalDataFrame
        real_date_column : str, optional
            Column name of the real date, by default "real_date"
        value_column : str, optional
            Column name of the value, by default "value"
        cid : Optional[str], optional
            `cid` to assign to the DataFrame, by default None. If not specified, the `cid`
            column must be present in the DataFrame.
        xcat : Optional[str], optional
            `xcat` to assign to the DataFrame, by default None
        ticker : Optional[str], optional
            Ticker to assign to the DataFrame, by default None
        categorical : bool, optional
            If True, convert the DataFrame to categorical, by default True

        Raises
        ------
        ValueError
            If the `real_date_column` or `value_column` are not found in the DataFrame,
            or if `ticker` is specified with `cid` or `xcat`.
        ValueError
            If the input DataFrame is empty.
        ValueError
            If the `cid` or `xcat` columns are not found in the DataFrame, and have not
            been specified in the function call.

        Returns
        -------
        QuantamentalDataFrame
            The QuantamentalDataFrame created from the long DataFrame.
        """

        if real_date_column not in df.columns:
            raise ValueError(f"No `{real_date_column}` column found in the DataFrame.")
        else:
            df = df.rename(columns={real_date_column: "real_date"})
            real_date_column = "real_date"

        if value_column not in df.columns:
            raise ValueError(f"No `{value_column}` column found in the DataFrame.")

        if len(df) == 0:
            raise ValueError("Input DataFrame is empty.")

        if ticker is not None:
            if bool(cid) or bool(xcat):
                raise ValueError("Cannot specify `ticker` with `cid` or `xcat`.")
            cid, xcat = ticker.split("_", 1)

        errstr = "No `{}` column found in the DataFrame and `{}` not specified."
        for var_, name_ in zip([cid, xcat], ["cid", "xcat"]):
            if name_ not in df.columns and var_ is None:
                raise ValueError(errstr.format(name_, name_))

        new_df = df[[real_date_column, value_column]].copy()

        # if the cid col is there in the df copy it over as a categorical type
        if "cid" in df.columns:
            new_df["cid"] = df["cid"].astype("category")
        # if the xcat col is there in the df copy it over as a categorical type
        if "xcat" in df.columns:
            new_df["xcat"] = df["xcat"].astype("category")

        # if the cid was specfied then overwrite and initialise as a single cid category dtype
        if cid is not None:
            new_df["cid"] = pd.Categorical.from_codes(
                codes=[0] * len(new_df), categories=[cid]
            )

        # if the xcat was specfied then overwrite and initialise as a single xcat category dtype
        if xcat is not None:
            new_df["xcat"] = pd.Categorical.from_codes(
                codes=[0] * len(new_df), categories=[xcat]
            )

        return QuantamentalDataFrame(new_df, categorical=categorical)

    @classmethod
    def from_qdf_list(
        cls,
        qdf_list: List["QuantamentalDataFrame"],
        categorical: bool = True,
    ) -> "QuantamentalDataFrame":
        """
        Concatenate a list of QuantamentalDataFrames into a single QuantamentalDataFrame.

        Parameters
        ----------
        qdf_list : List[QuantamentalDataFrame]
            List of QuantamentalDataFrames to concatenate
        categorical : bool, optional
            If True, convert the DataFrame to categorical, by default True

        Raises
        ------
        TypeError
            If any element in the list is not a QuantamentalDataFrame.
        ValueError
            If the input list is empty.

        Returns
        -------
        QuantamentalDataFrame
            The concatenated QuantamentalDataFrame.
        """
        if not all(isinstance(qdf, QuantamentalDataFrame) for qdf in qdf_list):
            raise TypeError("All elements in the list must be QuantamentalDataFrames.")
        if not qdf_list:
            raise ValueError("Input list is empty.")

        qdf_list = concat_qdfs(qdf_list)
        return QuantamentalDataFrame(qdf_list, categorical=categorical)

    @classmethod
    def from_wide(
        cls,
        df: pd.DataFrame,
        value_column: str = "value",
        categorical: bool = True,
    ) -> "QuantamentalDataFrame":
        """
        Convert a wide DataFrame to a QuantamentalDataFrame.

        Parameters
        ----------
        df : pd.DataFrame
            Wide DataFrame to convert to a QuantamentalDataFrame
        value_column : str, optional
            Column to pivot, by default "value"
        categorical : bool, optional
            If True, convert the DataFrame to categorical, by default True

        Raises
        ------
        TypeError
            If `df` is not a pandas DataFrame.
        ValueError
            If `df` does not have a datetime index.
        ValueError
            If all columns are not in the format 'cid_xcat'.

        Returns
        -------
        QuantamentalDataFrame
            The QuantamentalDataFrame created from the wide DataFrame.
        """

        if not isinstance(df, pd.DataFrame):
            raise TypeError("`df` must be a pandas DataFrame.")

        if not isinstance(df.index, pd.DatetimeIndex):
            raise ValueError("`df` must have a datetime index.")

        if not isinstance(value_column, str):
            raise TypeError("`value_column` must be a string.")

        if not all("_" in col for col in df.columns):
            raise ValueError("All columns must be in the format 'cid_xcat'.")

        df.index.name = "real_date"
        df.columns.name = None
        tickers = df.columns.tolist()
        qdfs_list: List[QuantamentalDataFrame] = []

        for tkr in tickers:
            cid, xcat = tkr.split("_", 1)
            qdf = df[[tkr]].reset_index().rename(columns={tkr: value_column})
            df = df.drop(columns=[tkr])
            qdf = QuantamentalDataFrame.from_long_df(
                df=qdf, cid=cid, xcat=xcat, value_column=value_column
            )
            qdfs_list.append(qdf)

        return QuantamentalDataFrame.from_qdf_list(qdfs_list, categorical=categorical)

    @classmethod
    def create_empty_df(
        cls,
        cid: Optional[str] = None,
        xcat: Optional[str] = None,
        ticker: Optional[str] = None,
        metrics: List[str] = ["value"],
        date_range: Optional[pd.DatetimeIndex] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        categorical: bool = True,
    ) -> "QuantamentalDataFrame":
        """
        Create an empty QuantamentalDataFrame.

        Parameters
        ----------
        cid : Optional[str], optional
            `cid` to assign to the DataFrame, by default None
        xcat : Optional[str], optional
            `xcat` to assign to the DataFrame, by default None
        ticker : Optional[str], optional
            Ticker to assign to the DataFrame, by default None. If specified, `cid` and
            `xcat` must not be specified.
        metrics : List[str], optional
            Metrics to assign to the DataFrame, by default ["value"]
        date_range : Optional[pd.DatetimeIndex], optional
            Date range to assign to the DataFrame, by default None. If not specified,
            `start` and `end` must be specified.
        start : Optional[str], optional
            Start date to assign to the DataFrame, by default None
        end : Optional[str], optional
            End date to assign to the DataFrame, by default None
        categorical : bool, optional
            If True, convert the DataFrame to categorical, by default True

        Returns
        -------
        QuantamentalDataFrame
            The empty QuantamentalDataFrame.
        """
        qdf = create_empty_categorical_qdf(
            cid=cid,
            xcat=xcat,
            ticker=ticker,
            metrics=metrics,
            date_range=date_range,
            start=start,
            end=end,
        )
        return QuantamentalDataFrame(qdf, categorical=categorical)



from collections import defaultdict
import numpy as np
import pandas as pd
import scipy.stats as stats

from sklearn.ensemble import VotingRegressor

from macrosynergy.learning.forecasting.model_systems import BaseRegressionSystem


def neg_mean_abs_corr(
    estimator,
    X_test,
    y_test,
    correlation_type="pearson",
):
    """
    Negative mean absolute correlation between a time series of benchmark returns and a
    panel of computed hedged returns, with average taken over all cross-sections.

    Parameters
    ----------
    estimator : BaseRegressionSystem
        A fitted `scikit-learn` regression object with separate linear models for each
        cross-section of returns, regressed against a time series of benchmark risk basket
        returns. It is expected to possess a `coefs_` dictionary attribute with keys
        corresponding to the cross-sections of returns and values corresponding to the
        estimated coefficients of the linear model for each cross-section.
    X_test : pd.DataFrame
        Risk-basket returns replicated for each cross-section of returns in `y_test`.
    y_test : pd.Series
        Panel of financial contract returns.
    correlation_type : str
        Type of correlation to compute between each hedged return
        series and the risk basket return series. Default is "pearson".
        Alternatives are "spearman" and "kendall".

    Returns
    -------
    neg_mean_abs_corr : float
        Negative mean absolute correlation between benchmark risk basket returns and
        computed hedged returns.

    Notes
    -----
    For each cross-section :math:`c` in `X_test`, hedged returns are calculated by
    subtracting :math:`X_{test, c} \\cdot \\text{coefs_}[c]` from each `y_{test, c}`.
    Following this, the negative mean absolute correlation over cross-sections can be
    calculated:

    ```{math}
    :label: neg_mean_abs_corr
    \\text{neg_mean_abs_corr} = - (1/C)\\sum_{c=1}^{C} \\left [ abs_corr_{c} \\right ]
    ```

    This function is a specialised scorer to evaluate the quality of a hedge within the
    `BetaEstimator` class in the `macrosynergy.learning` subpackage.
    """
    # Checks
    # estimator
    if isinstance(estimator, BaseRegressionSystem):
        if estimator.models_ is None:
            raise ValueError("estimator must be a fitted model.")
    elif isinstance(estimator, VotingRegressor):
        if not all(
            isinstance(est, BaseRegressionSystem) for est in estimator.estimators_
        ):
            raise TypeError(
                "estimator must be a VotingRegressor with BaseRegressionSystem estimators."
            )
        if not all(est.models_ is not None for est in estimator.estimators_):
            raise ValueError("estimator must be a VotingRegressor with fitted models.")
    else:
        raise TypeError(
            "estimator must be a BaseRegressionSystem or VotingRegressor object."
        )

    # X_test
    if not isinstance(X_test, pd.DataFrame):
        raise TypeError("X_test must be a pandas DataFrame.")
    if X_test.ndim != 2:
        raise ValueError("X_test must be a 2-dimensional DataFrame.")
    if X_test.shape[1] != 1:
        raise ValueError("X_test must have only one column.")
    if not isinstance(X_test.index, pd.MultiIndex):
        raise ValueError("X_test must be multi-indexed.")
    if not X_test.index.get_level_values(0).dtype == "object":
        raise TypeError("The outer index of X_test must be strings.")
    if not X_test.index.get_level_values(1).dtype == "datetime64[ns]":
        raise TypeError("The inner index of X_test must be datetime.date.")
    if not X_test.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
        raise ValueError(
            "The input feature matrix column for neg_mean_abs_corr",
            " must be numeric.",
        )
    if X_test.isnull().values.any():
        raise ValueError(
            "The input feature matrix for neg_mean_abs_corr must not contain any "
            "missing values."
        )

    # y_test
    if not isinstance(y_test, pd.Series):
        raise TypeError("y_test must be a pandas Series.")
    if not isinstance(y_test.index, pd.MultiIndex):
        raise ValueError("y_test must be multi-indexed.")
    if not y_test.index.get_level_values(0).dtype == "object":
        raise TypeError("The outer index of y_test must be strings.")
    if not y_test.index.get_level_values(1).dtype == "datetime64[ns]":
        raise TypeError("The inner index of y_test must be datetime.date.")
    if not y_test.index.equals(X_test.index):
        raise ValueError("y_test and X_test must have the same index.")
    if not pd.api.types.is_numeric_dtype(y_test):
        raise ValueError(
            "The input target vector for neg_mean_abs_corr",
            " must be numeric.",
        )
    if y_test.isnull().values.any():
        raise ValueError(
            "The input target vector for neg_mean_abs_corr must not contain any "
            "missing values."
        )
    # Obtain key information
    market_returns = X_test.iloc[:, 0].copy()
    contract_returns = y_test.copy()
    unique_cross_sections = X_test.index.get_level_values(0).unique()
    
    # Handle voting regressor case later
    if isinstance(estimator, VotingRegressor):
        estimators = estimator.estimators_
        coefs_list = [est.coefs_ for est in estimators]
        sum_dict = defaultdict(lambda: [0, 0])

        for coefs in coefs_list:
            for key, value in coefs.items():
                sum_dict[key][0] += value
                sum_dict[key][1] += 1

        estimated_coefs = {key: sum / count for key, (sum, count) in sum_dict.items()}
    else:
        estimated_coefs = estimator.coefs_

    running_sum = 0
    xs_count = 0
    for cross_section in unique_cross_sections:
        # Check whether a model for this cross-section has been estimated
        if cross_section in estimated_coefs.keys():
            xs_count += 1
            # Get cross-section returns and matched risk basket returns
            contract_returns_c = contract_returns.xs(cross_section)
            market_returns_c = market_returns.xs(cross_section)
            hedged_returns_c = (
                contract_returns_c - estimated_coefs[cross_section] * market_returns_c
            )
            # Compute negative absolute market correlation
            if correlation_type == "pearson":
                abs_corr = abs(stats.pearsonr(hedged_returns_c, market_returns_c)[0])
            elif correlation_type == "spearman":
                abs_corr = abs(stats.spearmanr(hedged_returns_c, market_returns_c)[0])
            else:
                # Use Kendall
                abs_corr = abs(stats.kendalltau(hedged_returns_c, market_returns_c)[0])
            # Update running sum
            running_sum += abs_corr
        else:
            # Then a model wasn't estimated for this cross-section
            continue

    if xs_count == 0:
        return np.nan
    else:
        return -running_sum / xs_count


if __name__ == "__main__":
    import macrosynergy.management as msm
    from macrosynergy.management.simulate import make_qdf
    from macrosynergy.learning import RidgeRegressionSystem

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "BMXR"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2012-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2012-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["BMXR"] = ["2012-01-01", "2020-12-31", 1, 2, 0.95, 1]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    Xy = msm.categories_df(
        df=dfd, xcats=xcats, cids=cids, freq="M", lag=1, xcat_aggs=["last", "sum"]
    ).dropna()
    X = Xy.iloc[:, :-1]
    y = Xy.iloc[:, -1]

    ridge = RidgeRegressionSystem()
    ridge.fit(X, y)
    print(
        "\nNegative mean absolute correlation: "
        f"{neg_mean_abs_corr(ridge, X, y, correlation_type='pearson')}"
    )



"""
Module for calculating z-scores for a panel around a neutral level ("zn scores").
"""

import numpy as np
import pandas as pd
from typing import List, Union
from numbers import Number
from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import (
    drop_nan_series,
    reduce_df,
    _map_to_business_day_frequency,
)
from macrosynergy.management.types import QuantamentalDataFrame
from numbers import Number


def make_zn_scores(
    df: pd.DataFrame,
    xcat: str,
    cids: List[str] = None,
    start: str = None,
    end: str = None,
    blacklist: dict = None,
    sequential: bool = True,
    min_obs: int = 261,
    iis: bool = True,
    neutral: Union[str, Number] = "zero",
    est_freq: str = "D",
    thresh: float = None,
    pan_weight: float = 1,
    postfix: str = "ZN",
    unscore: bool = False,
) -> pd.DataFrame:
    """
    Computes z-scores for a panel around a neutral level ("zn scores").

    Parameters
    ----------
    df : ~pandas.Dataframe
        standardized JPMaQS DataFrame with the necessary columns: 'cid', 'xcat',
        'real_date' and 'value'.
    xcat : str
        extended category for which the zn_score is calculated.
    cids : List[str]
        cross sections for which zn_scores are calculated; default is all available for
        category.
    start : str
        earliest date in ISO format. Default is None and earliest date in df is used.
    end : str
        latest date in ISO format. Default is None and latest date in df is used.
    blacklist : dict
        cross-sections with date ranges that should be excluded from the calculation of
        zn-scores. This means that not only are there no zn-score values calculated for
        these periods, but also that they are not used for the scoring of other periods.
    sequential : bool
        if True (default) score parameters (neutral level and mean absolute deviation)
        are estimated sequentially with concurrently available information only.
    min_obs : int
        the minimum number of observations required to calculate zn_scores. Default is
        261. The parameter is only applicable if the "sequential" parameter is set to True.
        Otherwise the neutral level and the mean absolute deviation are both computed in-
        sample and will use the full sample.
    iis : bool
        if True (default) zn-scores are also calculated for the initial sample period
        defined by min-obs on an in-sample basis to avoid losing history. This is irrelevant
        if sequential is set to False.
    neutral : str, Number
        method to determine neutral level. Default is 'zero'. Alternatives are 'mean',
        'median' or a number.
    est_freq : str
        the frequency at which mean absolute deviations or means are are re-estimated.
        The options are daily, weekly, monthly & quarterly: "D", "W", "M", "Q". Default is
        daily. Re-estimation is performed at period end.
    thresh : float
        threshold value beyond which scores are winsorized, i.e. contained at that
        threshold. The threshold is the maximum absolute score value that the function is
        allowed to produce. The minimum threshold is 1 mean absolute deviation.
    pan_weight : float
        weight of panel (versus individual cross section) for calculating the z-score
        parameters, i.e. the neutral level and the mean absolute deviation. Default is 1,
        i.e. panel data are the basis for the parameters. Lowest possible value is 0, i.e.
        parameters are all specific to cross section.
    postfix : str
        string appended to category name for output; default is "ZN".
    unscore : bool, default False
        If True, the function will apply the specified threshold to z-scores,
        but return values on the original scale. The `thresh` parameter will
        determine the z-score limits, and the winsorized values will be
        converted back to the original scale before being returned.

    Returns
    -------
    ~pandas.Dataframe
        standardized DataFrame with the zn-scores of the chosen category: 'cid', 'xcat',
        'real_date' and 'value'.


    .. note::
        The blacklist argument is a dictionary with cross-sections as keys and tuples of
        start and end dates of the blacklist periods in ISO formats as values. If one cross
        section has multiple blacklist periods, numbers are added to the keys (i.e. TRY_1,
        TRY_2, etc.)
    """

    expected_columns = ["cid", "xcat", "real_date", "value"]
    df = QuantamentalDataFrame(df[expected_columns])

    # --- Assertions
    err: str = (
        "The `neutral` parameter must be a number or a string with value,"
        " either 'mean', 'median' or 'zero'."
    )
    if not isinstance(neutral, Number):
        if not isinstance(neutral, str):
            raise TypeError(err)
        elif neutral not in ["mean", "median", "zero"]:
            raise ValueError(err)

    if thresh is not None:
        err: str = "The `thresh` parameter must a numerical value >= 1.0."
        if not isinstance(thresh, Number):
            raise TypeError(err)
        elif thresh < 1.0:
            raise ValueError(err)

    if not isinstance(iis, bool):
        raise TypeError("Parameter `iis` must be a boolean.")

    err = (
        "The `pan_weight` parameter must be a numerical value between 0 and 1 "
        "(inclusive)."
    )
    if not isinstance(pan_weight, Number):
        raise TypeError(err)
    elif not (0 <= pan_weight <= 1):
        raise ValueError(err)

    error_min = "Minimum observations must be a non-negative Integer value."
    if not isinstance(min_obs, int):
        raise TypeError(error_min)
    if min_obs < 0:
        raise ValueError(error_min)

    est_freq = _map_to_business_day_frequency(
        freq=est_freq, valid_freqs=["D", "W", "M", "Q"]
    )

    # --- Prepare re-estimation dates and time-series DataFrame.

    # Remove any additional metrics defined in the DataFrame.
    if cids is not None:
        missing_cids = set(cids).difference(set(df["cid"]))
        if missing_cids:
            raise ValueError(
                f"The following cids are not available in the DataFrame: "
                f"{missing_cids}."
            )
    if xcat not in df["xcat"].unique():
        raise ValueError(f"The xcat {xcat} is not available in the DataFrame.")

    df = reduce_df(
        df, xcats=[xcat], cids=cids, start=start, end=end, blacklist=blacklist
    )

    if df.isna().values.any():
        df = drop_nan_series(df=df, raise_warning=True)

    s_date = min(df["real_date"])
    e_date = max(df["real_date"])
    dates_iter = pd.date_range(start=s_date, end=e_date, freq=est_freq)

    dfw = df.pivot(index="real_date", columns="cid", values="value")
    cross_sections = dfw.columns

    # --- The actual scoring.

    dfw_zns_pan = dfw * 0
    dfw_zns_css = dfw * 0

    if dfw.shape[0] < min_obs and pan_weight < 1 and pan_weight > 0:
        raise ValueError(
            f"The DataFrame has less than {min_obs} observations. "
            "Please adjust the `min_obs` parameter."
        )

    dfx_pan, df_mabs_pan, df_neutral_pan = None, None, None
    if pan_weight > 0:
        df_neutral_pan = expanding_stat(
            dfw,
            dates_iter,
            stat=neutral,
            sequential=sequential,
            min_obs=min_obs,
            iis=iis,
        )
        dfx_pan = dfw.sub(df_neutral_pan["value"], axis=0)
        df_mabs_pan = expanding_stat(
            dfx_pan.abs(),
            dates_iter,
            stat="mean",
            sequential=sequential,
            min_obs=min_obs,
            iis=iis,
        )
        dfw_zns_pan = dfx_pan.div(df_mabs_pan["value"], axis="rows")

    cid_dfx, cid_mabs, cid_neutral = {}, {}, {}
    if pan_weight < 1:
        for cid in cross_sections:
            dfi = dfw[cid]

            df_neutral = expanding_stat(
                dfi.to_frame(name=cid),
                dates_iter,
                stat=neutral,
                sequential=sequential,
                min_obs=min_obs,
                iis=iis,
            )
            dfx = dfi - df_neutral["value"]

            df_mabs = expanding_stat(
                dfx.abs().to_frame(name=cid),
                dates_iter,
                stat="mean",
                sequential=sequential,
                min_obs=min_obs,
                iis=iis,
            )
            dfx = pd.DataFrame(data=dfx.to_numpy(), index=dfx.index, columns=["value"])
            dfx = dfx.rename_axis("cid", axis=1)

            zns_css_df = dfx / df_mabs
            dfw_zns_css.loc[:, cid] = zns_css_df["value"]

            cid_dfx[cid] = dfx
            cid_mabs[cid] = df_mabs["value"]
            cid_neutral[cid] = df_neutral["value"]

    dfw_zns = (dfw_zns_pan * pan_weight) + (dfw_zns_css * (1 - pan_weight))

    dfw_zns = dfw_zns.dropna(axis=0, how="all")

    if thresh is not None:
        dfw_zns.clip(lower=-thresh, upper=thresh, inplace=True)

    if unscore:
        dfw_zns = _unscore_dfw_zns(
            dfw_zns,
            dfw_zns_pan,
            dfw_zns_css,
            df_mabs_pan,
            df_neutral_pan,
            cid_mabs,
            cid_neutral,
            cross_sections,
            pan_weight,
        )

    # --- Reformatting of output into standardised DataFrame.

    df_out = dfw_zns.stack().to_frame("value").reset_index()
    df_out = QuantamentalDataFrame.from_long_df(
        df=df_out,
        xcat=xcat + postfix,
        categorical=df.InitializedAsCategorical,
    )

    return df_out


def expanding_stat(
    df: pd.DataFrame,
    dates_iter: pd.DatetimeIndex,
    stat: Union[str, Number] = "mean",
    sequential: bool = True,
    min_obs: int = 261,
    iis: bool = True,
) -> pd.DataFrame:
    """
    Compute specified statistic based on an expanding sample.

    Parameters
    ----------
    df : ~pandas.Dataframe
        Daily-frequency time series DataFrame.
    dates_iter : ~pandas.DatetimeIndex
        controls the frequency of the neutral & mean absolute deviation calculations.
    stat : str, Number
        statistical method to be applied. This is typically 'mean', or 'median'.
    sequential : bool
        if True (default) the statistic is estimated sequentially. If this set to false
        a single value is calculated per time series, based on the full sample.
    min_obs : int
        minimum required observations for calculation of the statistic in days.
    iis : bool
        if set to True, the values of the initial interval determined by min_obs will be
        estimated in-sample, based on the full initial sample.

    Returns
    -------
    ~pandas.DataFrame
        Time series dataframe of the chosen statistic across all columns
    """

    df_out = pd.DataFrame(np.nan, index=df.index, columns=["value"])
    # An adjustment for individual series' first realised value is not required given the
    # returned DataFrame will be subtracted from the original DataFrame. The original
    # DataFrame will implicitly host this information through NaN values such that when
    # the arithmetic operation is made, any falsified values will be displaced by NaN
    # values.

    first_observation = df.dropna(axis=0, how="all").index[0]
    # Adjust for individual cross-sections' series commencing at different dates.
    first_estimation = df.dropna(axis=0, how="all").index[min_obs]

    obs_index = np.where(df.index == first_observation)[0][0]
    est_index = np.where(df.index == first_estimation)[0][0]

    if stat == "zero":
        df_out["value"] = 0

    elif isinstance(stat, Number):
        df_out["value"] = stat

    elif not sequential:
        # The entire series is treated as in-sample. Will automatically handle NaN
        # values.
        statval = df.stack().apply(stat)
        df_out["value"] = statval

    else:
        dates = dates_iter[dates_iter >= first_estimation]
        if stat == "mean":
            expanding_count = _get_expanding_count(
                df.loc[first_observation:], min_periods=min_obs + 1
            )
            df_mean = (
                df.loc[first_observation:]
                .sum(1)
                .expanding(min_periods=min_obs + 1)
                .sum()
                / expanding_count
            )
            try:
                df_mean = df_mean.dropna().loc[dates]
            except KeyError as e:
                err_str = 'Some dates in "dates_iter" have no corresponding data.'
                raise KeyError(err_str) from e

            df_mean.name = "value"
            df_out.update(df_mean)
        else:
            for date in dates:
                df_out.loc[date, "value"] = (
                    df.loc[first_observation:date].stack().apply(stat)
                )

        df_out = df_out.ffill()

        if iis and (est_index - obs_index) > 0:
            df_out = df_out.bfill(limit=int(est_index - obs_index))

    df_out.columns.name = "cid"
    return df_out


def _get_expanding_count(X: pd.DataFrame, min_periods: int = 1):
    """
    Helper method to get the number of non-NaN values in each expanding window.

    Parameters
    ----------
    X : ~pandas.DataFrame
        Pandas dataframe of input features.
    min_periods : int
        Minimum number of observations in window required to have a value (otherwise
        result is 0.).

    Returns
    -------
    ~numpy.ndarray
        Numpy array of expanding counts.
    """

    return X.expanding(min_periods).count().sum(1).to_numpy()


def _unscore_dfw_zns(
    dfw_zns: pd.DataFrame,
    dfw_zns_pan: pd.DataFrame,
    dfw_zns_css: pd.DataFrame,
    df_mabs_pan: pd.DataFrame,
    df_neutral_pan: pd.DataFrame,
    cid_mabs: dict,
    cid_neutral: dict,
    cross_sections: list,
    pan_weight: float,
) -> pd.DataFrame:
    """
    Unscore the weighted panel and cross-sectional components of dfw_zns.

    Parameters
    ----------
    dfw_zns : pd.DataFrame
        The combined z-scored DataFrame.
    dfw_zns_pan : pd.DataFrame
        The panel component of dfw_zns.
    dfw_zns_css : pd.DataFrame
        The cross-sectional component of dfw_zns.
    df_mabs_pan : pd.DataFrame
        Mean absolute deviation for the panel component.
    df_neutral_pan : pd.DataFrame
        Neutral component for the panel component.
    cid_mabs : dict
        Dictionary of mean absolute deviations per cross-section.
    cid_neutral : dict
        Dictionary of neutral components per cross-section.
    cross_sections : list
        List of cross-section identifiers.
    pan_weight : float
        The weight of the panel component, ranging from 0 to 1.

    Returns
    -------
    pd.DataFrame
        The unscored DataFrame.
    """

    if pan_weight > 0:
        dfw_zns_pan = (dfw_zns - (dfw_zns_css * (1 - pan_weight))) / pan_weight
        dfw_unscored_pan = dfw_zns_pan.mul(df_mabs_pan["value"], axis=0).add(
            df_neutral_pan["value"], axis=0
        )
    else:
        dfw_unscored_pan = pd.DataFrame(
            0, index=dfw_zns.index, columns=dfw_zns.columns
        )

    if pan_weight < 1:
        dfw_zns_css = (dfw_zns - (dfw_zns_pan * pan_weight)) / (1 - pan_weight)
        dfw_unscored_css = pd.DataFrame(
            index=dfw_zns.index, columns=dfw_zns.columns
        )
        for cid in cross_sections:
            dfw_unscored_css[cid] = (
                dfw_zns_css[cid] * cid_mabs[cid]
            ) + cid_neutral[cid]
    else:
        dfw_unscored_css = pd.DataFrame(
            0, index=dfw_zns.index, columns=dfw_zns.columns
        )

    if pan_weight == 1:
        dfw_unscored = dfw_unscored_pan
    elif pan_weight == 0:
        dfw_unscored = dfw_unscored_css
    else:
        dfw_unscored = dfw_unscored_css
    dfw_zns = dfw_unscored

    return dfw_unscored

if __name__ == "__main__":
    np.random.seed(1)

    cids = ["AUD", "CAD", "GBP", "USD", "NZD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )

    df_cids.loc["AUD"] = ["2010-01-01", "2020-12-31", 0.5, 2]
    df_cids.loc["CAD"] = ["2006-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP"] = ["2008-01-01", "2020-11-30", -0.2, 0.5]
    df_cids.loc["USD"] = ["2007-01-01", "2020-09-30", -0.2, 0.5]
    df_cids.loc["NZD"] = ["2002-01-01", "2020-09-30", -0.1, 2]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR"] = ["2008-01-01", "2020-12-31", 0, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]
    df_xcats.loc["GROWTH"] = ["2012-01-01", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2013-01-01", "2020-10-30", 1, 2, 0.8, 0.5]

    # Apply a blacklist period from series' start date.
    black = {"AUD": ["2010-01-01", "2013-12-31"], "GBP": ["2018-01-01", "2100-01-01"]}

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])

    # Monthly: panel + cross.
    dfzm = make_zn_scores(
        dfd.copy(deep=True),
        xcat="XR",
        sequential=True,
        cids=cids,
        blacklist=black,
        iis=True,
        neutral="mean",
        pan_weight=0.5,
        min_obs=261,
        est_freq="D",
        unscore=True,
        # thresh=5
    )
    print(dfzm)

    # Weekly: panel + cross.
    dfzw = make_zn_scores(
        dfd,
        xcat="XR",
        sequential=True,
        cids=cids,
        blacklist=black,
        iis=False,
        neutral="mean",
        pan_weight=0.5,
        min_obs=261,
        est_freq="w",
    )

    # Daily: panel. Neutral and mean absolute deviation will be computed daily.
    dfzd = make_zn_scores(
        dfd,
        xcat="XR",
        sequential=True,
        cids=cids,
        blacklist=black,
        iis=True,
        neutral="mean",
        pan_weight=1.0,
        min_obs=261,
        est_freq="d",
    )

    # Daily: cross.
    dfd["ticker"] = dfd["cid"] + "_" + dfd["xcat"]
    dfzd = make_zn_scores(
        dfd,
        xcat="XR",
        sequential=True,
        cids=cids,
        blacklist=black,
        iis=True,
        neutral="mean",
        pan_weight=0.0,
        min_obs=261,
        est_freq="d",
    )

    panel_df = make_zn_scores(
        dfd,
        "CRY",
        cids,
        start="2010-01-04",
        blacklist=black,
        sequential=False,
        min_obs=0,
        neutral="mean",
        iis=True,
        thresh=None,
        pan_weight=0.75,
        postfix="ZN",
    )

    print(panel_df)

    panel_df_7 = make_zn_scores(
        dfd,
        "CRY",
        cids,
        start="2010-01-04",
        blacklist=black,
        sequential=False,
        min_obs=0,
        neutral="zero",
        iis=True,
        thresh=None,
        pan_weight=0.75,
        postfix="ZN",
    )

    print(panel_df_7)



"""
Class to estimate market betas and calculate out-of-sample hedged returns based on
sequential learning. 
"""

from collections import defaultdict

import numpy as np
import pandas as pd
from sklearn.ensemble import VotingRegressor

from macrosynergy.learning.forecasting.model_systems.base_regression_system import (
    BaseRegressionSystem,
)
from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy.learning import ExpandingFrequencyPanelSplit
from macrosynergy.learning.sequential import BasePanelLearner
from macrosynergy.management import categories_df, reduce_df, update_df
from macrosynergy.management.utils.df_utils import (
    concat_categorical,
    _insert_as_categorical,
)


class BetaEstimator(BasePanelLearner):
    """
    Class for sequential beta estimation by learning optimal regression coefficients.
    Out-of-sample hedged returns are additionally calculated and stored.

    Parameters
    ----------
    df : pd.DataFrame
        Daily quantamental dataframe in JPMaQS format containing a panel of features, as
        well as a panel of returns.
    xcats : str or list
        Name of a market return category within the panel specified in `df`.
    benchmark_return : str
        Name of the benchmark return ticker within the panel specified in `df`.
    cids : list, optional
        List of cross-sections for which hedged returns are to be calculated.
        Default is None, which calculates hedged returns for all cross-sections in the
        return panel.
    start : str, optional
        Start date for considered data in subsequent analysis in ISO 8601 format.
        Default is None i.e. the earliest date in the dataframe.
    end : str, optional
        End date for considered data in subsequent analysis in ISO 8601 format.
        Default is None i.e. the latest date in the dataframe.

    Notes
    -----
    The `BetaEstimator` class is used to sequentially estimate macro betas based on a
    panel of contract returns (provided in `xcats`) and a benchmark return ticker
    (provided in `benchmark_return`). The initial conditions of the learning process
    are specified by the dimensions of an initial training set. An optimal model is
    selected out of a provided collection (with associated hyperparameters), a beta is
    extracted for each cross-section (subject to availability) and out-of-sample hedged
    returns are calculated for each cross-section with an estimated beta. The betas and
    hedged returns are stored, and the training set is expanded to include the now-realized
    returns. This process is repeated until the end of the dataset is reached.

    In addition to storing betas and hedged returns, this class also stores useful model
    selection information for analysis, such as the models selected at each point in time.

    Model and hyperparameter selection is performed by cross-validation. Given a
    collection of models and associated hyperparameters to choose from, an HPO is run
    - currently only grid search and random search are supported - to determine the
    optimal choice. This is done by providing a collection of `scikit-learn` compatible
    scoring functions, as well as a collection of `scikit-learn` compatible
    cross-validation splitters and scorers. At each point in time, the cross-validation
    folds are the union of the folds produced by each splitter provided. Each scorer is
    evaluated on each test fold and summarised across test folds by either a custom
    function provided by the user or a common string i.e. 'mean'.

    Consequently, each model and hyperparameter combination has an associated collection
    of scores induced by different metrics, in units of those scorers. In order to form a
    composite score for each hyperparameter, the scores must be normalized across
    model/hyperparameter combinations. This makes scores across scorers comparable, so
    that the average score across adjusted scores can be used as a meaningful estimate
    of each model's generalization ability. Finally, a composite score for each model and
    hyperparameter combination is calculated by averaging the adjusted scores across all
    scorers.

    The optimal model is the one with the largest composite score.
    """

    def __init__(
        self,
        df,
        xcats,
        benchmark_return,
        cids=None,
        start=None,
        end=None,
    ):
        # Checks
        # TODO: Refactor these checks.
        if not isinstance(df, pd.DataFrame):
            raise TypeError("df must be a pandas DataFrame.")
        if not set(["cid", "xcat", "real_date", "value"]).issubset(df.columns):
            raise ValueError(
                "df must have columns 'cid', 'xcat', 'real_date' and 'value'."
            )

        # cids checks
        if cids is not None:
            if not isinstance(cids, list):
                raise TypeError("cids must be a list.")
            if not all(isinstance(cid, str) for cid in cids):
                raise TypeError("All elements in cids must be strings.")
            for cid in cids:
                if cid not in df["cid"].unique():
                    raise ValueError(f"{cid} not in the dataframe.")

        if not isinstance(benchmark_return, str):
            raise TypeError("benchmark_return must be a string.")

        if isinstance(xcats, str):
            xcats = [xcats]
        elif isinstance(xcats, list):
            if not all(isinstance(xcat, str) for xcat in xcats):
                raise TypeError("All elements in xcats must be strings.")
            elif len(xcats) != 1:
                raise ValueError("xcats must be a string or a list of a single xcat.")
        else:
            raise TypeError("xcats must be a string or a list of a single xcat.")

        # Create pseudo-panel
        dfx = pd.DataFrame(columns=["real_date", "cid", "xcat", "value"])

        self.benchmark_return = benchmark_return
        self.benchmark_cid, self.benchmark_xcat = benchmark_return.split("_", 1)

        for cid in cids:
            # Extract cross-section contract returns
            dfa = reduce_df(
                df=df,
                xcats=xcats if isinstance(xcats, list) else [xcats],
                cids=[cid],
            )
            # Extract benchmark returns
            dfb = reduce_df(
                df=df,
                xcats=[self.benchmark_xcat],
                cids=[self.benchmark_cid],
            )

            # Combine contract and benchmark returns and rename cross-section identifier
            # in order to match the benchmark return with each cross section in a pseudo
            # panel
            df_cid = pd.concat([dfa, dfb], axis=0)
            df_cid["cid"] = f"{cid}v{self.benchmark_cid}"

            dfx = update_df(dfx, df_cid)

        super().__init__(
            df=dfx,
            xcats=(
                [self.benchmark_xcat] + xcats
                if isinstance(xcats, list)
                else [self.benchmark_xcat, xcats]
            ),
            cids=list(dfx["cid"].unique()),
            start=start,
            end=end,
            blacklist=None,
            freq="D",
            lag=0,
        )

        # Create forecast dataframe index
        min_date = min(self.unique_date_levels)
        max_date = max(self.unique_date_levels)
        forecast_date_levels = pd.date_range(start=min_date, end=max_date, freq="B")
        self.forecast_idxs = pd.MultiIndex.from_product(
            [
                [cid.split("v")[0] for cid in self.unique_xs_levels],
                forecast_date_levels,
            ],
            names=["cid", "real_date"],
        )

        # Create initial dataframes to store estimated betas and OOS hedged returns
        self.betas = pd.DataFrame(columns=["real_date", "cid", "xcat", "value"]).astype(
            {
                "real_date": "datetime64[s]",
                "cid": "category",
                "xcat": "category",
                "value": "float32",
            }
        )
        self.hedged_returns = pd.DataFrame(
            columns=["real_date", "cid", "xcat", "value"]
        ).astype(
            {
                "real_date": "datetime64[s]",
                "cid": "category",
                "xcat": "category",
                "value": "float32",
            }
        )

    def estimate_beta(
        self,
        beta_xcat,
        hedged_return_xcat,
        models,
        hyperparameters,
        scorers,
        inner_splitters,
        search_type="grid",
        normalize_fold_results=False,
        cv_summary="mean",
        min_cids=4,
        min_periods=12 * 3,
        est_freq="D",
        max_periods=None,
        split_functions=None,
        n_iter=None,
        n_jobs_outer=-1,
        n_jobs_inner=1,
    ):
        """
        Determines optimal model betas and associated out-of-sample hedged returns.

        Parameters
        ----------
        beta_xcat : str
            Category name for the panel of estimated betas.
        hedged_return_xcat : str
            Category name for the panel of out-of-sample hedged returns.
        models : dict
            Dictionary of models to choose from. The keys are model names and the values
            are scikit-learn compatible models.
        hyperparameters : dict
            Dictionary of hyperparameters to choose from. The keys are model names and
            the values are hyperparameter dictionaries for the corresponding model. The
            keys must match with those provided in `models`.
        scorers : dict
            Dictionary of scoring functions to use in the hyperparameter optimization
            process. The keys are scorer names and the values are scikit-learn compatible
            scoring functions.
        inner_splitters : dict
            Dictionary of inner splitters to use in the hyperparameter optimization
            process. The keys are splitter names and the values are scikit-learn compatible
            cross-validator objects.
        search_type : str
            Type of hyperparameter optimization to perform. Default is "grid". Options are
            "grid" and "prior".
        normalize_fold_results : bool
            Whether to normalize the scores across folds before combining them. Default is
            False.
        cv_summary : str or callable
            Summary function to use to combine scores across cross-validation folds.
            Default is "mean". Options are "mean", "median" or a callable function.
        min_cids : int
            Minimum number of cross-sections required for the initial
            training set. Default is 4.
        min_periods : int
            Minimum number of periods required for the initial training set, in units of
            the frequency `freq` specified in the constructor. Default is 36.
        est_freq : str
            Frequency at which models are refreshed. This corresponds with forward
            frequency of out-of-sample hedged returns and the frequency at which betas
            are estimated.
        max_periods : int
            Maximum length of each training set in units of the frequency `freq` specified
            in the constructor. Default is None, in which case the sequential optimization
            uses expanding training sets, as opposed to rolling windows.
        split_functions : dict, optional
            Dict of callables for determining the number of cross-validation
            splits to add to the initial number as a function of the number of iterations
            passed in the sequential learning process. Default is None. The keys must
            correspond to the keys in `inner_splitters` and should be set to None for any
            splitters that do not require splitter adjustment.
        n_iter : int, optional
            Number of iterations to run in random hyperparameter search. Default is None.
        n_jobs_outer : int, optional
            Number of jobs to run in parallel for the outer sequential loop. Default is -1.
            It is advised for n_jobs_inner * n_jobs_outer (replacing -1 with the number of
            available cores) to be less than or equal to the number of available cores on
            the machine.
        n_jobs_inner : int, optional
            Number of jobs to run in parallel for the inner loop. Default is 1.
            It is advised for n_jobs_inner * n_jobs_outer (replacing -1 with the number of
            available cores) to be less than or equal to the number of available cores on
            the machine.
        """
        # Checks
        # All others are checked in the run method
        if not isinstance(hedged_return_xcat, str):
            raise TypeError("hedged_return_xcat must be a string.")

        self.hedged_return_xcat = hedged_return_xcat

        # Set up outer splitter
        outer_splitter = ExpandingFrequencyPanelSplit(
            expansion_freq=est_freq,
            test_freq=est_freq,
            min_cids=min_cids,
            min_periods=min_periods,
        )

        # Run pipeline
        results = self.run(
            name=beta_xcat,
            outer_splitter=outer_splitter,
            inner_splitters=inner_splitters,
            models=models,
            hyperparameters=hyperparameters,
            scorers=scorers,
            search_type=search_type,
            normalize_fold_results=normalize_fold_results,
            cv_summary=cv_summary,
            split_functions=split_functions,
            n_iter=n_iter,
            n_jobs_outer=n_jobs_outer,
            n_jobs_inner=n_jobs_inner,
        )

        if hedged_return_xcat in self.hedged_returns["xcat"].unique():
            self.hedged_returns = self.hedged_returns[
                self.hedged_returns.xcat != hedged_return_xcat
            ]
        if beta_xcat in self.betas["xcat"].unique():
            self.betas = self.betas[self.betas.xcat != beta_xcat]
        if beta_xcat in self.chosen_models.name.unique():
            self.chosen_models = self.chosen_models[
                self.chosen_models.name != beta_xcat
            ]

        # Collect results from the worker
        beta_data = []
        hedged_return_data = []
        model_choice_data = []

        for split_result in results:
            beta_data.extend(split_result["betas"])
            hedged_return_data.extend(split_result["hedged_returns"])
            model_choice_data.append(split_result["model_choice"])

        stored_betas = pd.DataFrame(
            index=self.forecast_idxs, columns=[beta_xcat], data=np.nan, dtype="float32"
        )
        # Create quantamental dataframes of betas and hedged returns
        for real_date, cid, value in beta_data:
            stored_betas.loc[(cid, real_date), beta_xcat] = value

        stored_betas = stored_betas.groupby(level=0, observed=True).ffill().dropna()
        stored_betas.columns = stored_betas.columns.astype("category")
        stored_betas_long = pd.melt(
            frame=stored_betas.reset_index(),
            id_vars=["real_date", "cid"],
            var_name="xcat",
            value_name="value",
        )

        hedged_returns = (
            pd.DataFrame(hedged_return_data, columns=["real_date", "cid", "value"])
            .sort_values(["cid", "real_date"])
            .dropna()
        ).astype({"cid": "category"})
        hedged_returns = _insert_as_categorical(
            hedged_returns, "xcat", hedged_return_xcat, 2
        )

        self.betas = concat_categorical(self.betas, stored_betas_long)
        self.hedged_returns = concat_categorical(
            self.hedged_returns,
            hedged_returns,
        )

        # Store model selection data
        model_df_long = pd.DataFrame(
            columns=[col for col in self.chosen_models.columns if col != "name"],
            data=model_choice_data,
        ).astype({"model_type": "category"})
        model_df_long = _insert_as_categorical(model_df_long, "name", beta_xcat, 1)

        self.chosen_models = concat_categorical(self.chosen_models, model_df_long)

    def store_split_data(
        self,
        pipeline_name,
        optimal_model,
        optimal_model_name,
        optimal_model_score,
        optimal_model_params,
        inner_splitters_adj,
        X_train,
        y_train,
        X_test,
        y_test,
        timestamp,
        adjusted_test_index,
    ):
        """
        Stores characteristics of the optimal model at each retraining date.

        Parameters
        ----------
        pipeline_name : str
            Name of the signal optimization process.
        optimal_model : BaseRegressionSystem or VotingRegressor
            Optimal model selected at each retraining date.
        optimal_model_name : str
            Name of the optimal model.
        optimal_model_score : float
            Cross-validation score for the optimal model.
        optimal_model_params : dict
            Chosen hyperparameters for the optimal model.
        inner_splitters_adj : dict
            Dictionary of adjusted inner splitters.
        X_train : pd.DataFrame
            Training feature matrix.
        y_train : pd.Series
            Training response variable.
        X_test : pd.DataFrame
            Test feature matrix.
        y_test : pd.Series
            Test response variable.
        timestamp : pd.Timestamp
            Timestamp of the retraining date.
        adjusted_test_index : pd.MultiIndex
            Adjusted test index to account for lagged features.

        Returns
        -------
        dict
            Dictionary containing the betas and hedged returns determined at the
            given retraining date.
        """
        if isinstance(optimal_model, VotingRegressor):
            estimators = optimal_model.estimators_
            coefs_list = [est.coefs_ for est in estimators]
            sum_dict = defaultdict(lambda: [0, 0])

            for coefs in coefs_list:
                for key, value in coefs.items():
                    sum_dict[key][0] += value
                    sum_dict[key][1] += 1
            betas = {key: sum / count for key, (sum, count) in sum_dict.items()}
        elif isinstance(optimal_model, BaseRegressionSystem):
            betas = optimal_model.coefs_
        else:
            X_train.index.get_level_values(0).unique()
            betas = {cid: np.nan for cid in X_train.index.get_level_values(0).unique()}

        betas_list = [
            [
                X_train.index.get_level_values(1).max(),
                cid.split("v")[0],
                beta,
            ]
            for cid, beta in betas.items()
        ]

        # Now calculate the induced hedged returns

        betas_series = pd.Series(betas)
        XB = X_test.mul(betas_series, level=0, axis=0)
        hedged_returns = y_test.values.reshape(-1, 1) - XB.values.reshape(-1, 1)
        hedged_returns_data = [
            [idx[1], idx[0].split("v")[0]] + [hedged_returns[i].item()]
            for i, (idx, _) in enumerate(y_test.items())
        ]
        return {"betas": betas_list, "hedged_returns": hedged_returns_data}

    def evaluate_hedged_returns(
        self,
        hedged_return_xcat=None,
        cids=None,
        correlation_types="pearson",
        title=None,
        start=None,
        end=None,
        blacklist=None,
        freqs="M",
    ):
        """
        Method to determine and display a table of average absolute correlations between
        the benchmark return and the computed hedged returns within the class instance, over
        all cross-sections in the panel. Additionally, the correlation table displays the
        same results for the unhedged return specified in the class instance for comparison
        purposes.

        The returned dataframe will be multi-indexed by (benchmark return, return, frequency)
        and will contain each computed absolute correlation coefficient on each column.

        Parameters
        ----------
        hedged_return_xcat : str or list, optional
            Hedged returns to be evaluated. Default is None, which evaluates all hedged
            returns within the class instance.
        cids : str or list, optional
            Cross-sections for which evaluation of hedged returns takes place.
            Default is None, which evaluates all cross-sections within the class instance.
        correlation_types : str or list, optional
            Types of correlations to calculate.
            Options are "pearson", "spearman" and "kendall". If None, all three
            are calculated. Default is "pearson".
        title : str, optional
            Title for the correlation table. If None, the default
            title is "Average absolute correlations between each return and the chosen
            benchmark". Default is None.
        start : str, optional
            String in ISO format. Default is None.
        end : str, optional
            String in ISO format. Default is None.
        blacklist : dict, optional
            Dictionary of tuples of start and end dates to exclude from the evaluation.
            Default is None.
        freqs: str or list, optional
            Letters denoting all frequencies at which the correlations may be calculated.
            This must be a selection of "D", "W", "M", "Q" and "A". Default is "M".
            Each return series will always be summed over the sample period.

        Returns
        -------
        pd.DataFrame
            A dataframe of average absolute correlations between the benchmark return and the
            computed hedged returns.
        """
        # Checks
        correlation_types, hedged_return_xcat, cids, freqs = (
            self._checks_evaluate_hedged_returns(
                correlation_types=correlation_types,
                hedged_return_xcat=hedged_return_xcat,
                cids=cids,
                start=start,
                end=end,
                blacklist=blacklist,
                freqs=freqs,
            )
        )

        cids_v_benchmark = [f"{cid}v{self.benchmark_cid}" for cid in cids]
        # Construct a quantamental dataframe comprising specified hedged returns as well
        # as the unhedged returns and the benchmark return specified in the class instance
        hedged_df = self.hedged_returns[
            (self.hedged_returns["xcat"].isin(hedged_return_xcat))
            & (self.hedged_returns["cid"].isin(cids))
        ]
        unhedged_df = self.df[
            (self.df["xcat"].isin(self.xcats)) & (self.df["cid"].isin(cids_v_benchmark))
        ]
        benchmark_df = self.df[
            (self.df["xcat"] == self.benchmark_xcat)
            & (self.df["cid"] == f"{self.benchmark_cid}v{self.benchmark_cid}")
        ]

        cid_mapping = dict(zip(cids, cids_v_benchmark))
        hedged_df["cid"] = hedged_df["cid"].replace(cid_mapping)
        combined_df = concat_categorical(hedged_df, unhedged_df)

        # Create a pseudo-panel to match contract return cross-sections with a replicated
        # benchmark return. This is multi-indexed by (new cid, real_date). The columns
        # are the named hedged returns, with the final column being the benchmark category.
        dfx = pd.DataFrame(columns=["real_date", "cid", "xcat", "value"])

        for cid in cids_v_benchmark:
            # Extract unhedged and hedged returns
            dfa = reduce_df(
                df=combined_df,
                xcats=hedged_return_xcat + self.xcats,
                cids=[cid],
            )
            # Extract benchmark returns
            dfb = reduce_df(
                df=benchmark_df,
                xcats=[self.benchmark_xcat],
                cids=[self.benchmark_cid],
            )
            # Combine and rename cross-section
            df_cid = concat_categorical(dfa, dfb)

            dfx = update_df(dfx, df_cid)

        # Create long format dataframes for each specified frequency
        Xy_long_freq = []
        for freq in freqs:
            Xy_long = categories_df(
                df=dfx,
                xcats=hedged_return_xcat + self.xcats,
                cids=[cid for cid in cids_v_benchmark],
                start=start,
                end=end,
                blacklist=blacklist,
                freq=freq,
                xcat_aggs=["sum", "sum"],
            )
            Xy_long_freq.append(Xy_long)

        # For each xcat and frequency, calculate the mean absolute correlations
        # between the benchmark return and the (hedged and unhedged) market returns
        df_rows = []
        xcats_non_benchmark = [
            xcat for xcat in self.xcats if xcat != self.benchmark_xcat
        ]
        for xcat in hedged_return_xcat + xcats_non_benchmark:
            for freq, Xy_long in zip(freqs, Xy_long_freq):
                calculated_correlations = []
                for correlation in correlation_types:
                    calculated_correlations.append(
                        self._get_mean_abs_corrs(
                            xcat=xcat,
                            df=Xy_long,
                            correlation=correlation,
                            cids=cids,
                        )
                    )
                df_rows.append(calculated_correlations)
        # Create underlying dataframe to store the results
        multiindex = pd.MultiIndex.from_product(
            [[self.benchmark_return], hedged_return_xcat + xcats_non_benchmark, freqs],
            names=["benchmark return", "return category", "frequency"],
        )
        corr_df = pd.DataFrame(
            columns=[correlation for correlation in correlation_types],
            index=multiindex,
            data=df_rows,
        )

        return corr_df

    def _checks_evaluate_hedged_returns(
        self,
        correlation_types,
        hedged_return_xcat,
        cids,
        start,
        end,
        blacklist,
        freqs,
    ):
        """
        Input checks for the `evaluate_hedged_returns()` method.

        Parameters
        ----------
        correlation_types : str or list
            Types of correlations to calculate.
        hedged_return_xcat : str or list, optional
            Hedged returns to be evaluated.
        cids : str or list, optional
            Cross-sections for which evaluation of hedged returns takes place.
        start : str, optional
            Start date for evaluation.
        end : str, optional
            End date for evaluation.
        blacklist : dict, optional
            Dictionary of tuples of start and end dates to exclude from the evaluation.
        freqs: str or list, optional
            Letters denoting all frequencies at which the correlations may be calculated.
        """
        if isinstance(correlation_types, str):
            correlation_types = [correlation_types]
        elif not isinstance(correlation_types, list):
            raise TypeError("correlation_types must be a string or a list")
        if not all(
            isinstance(correlation_type, str) for correlation_type in correlation_types
        ):
            raise TypeError("All elements in correlation_types must be strings.")
        if not all(
            correlation_type in ["pearson", "spearman", "kendall"]
            for correlation_type in correlation_types
        ):
            raise ValueError(
                "All elements in correlation_types must be one of 'pearson', 'spearman' or 'kendall'."
            )

        if hedged_return_xcat is None:
            hedged_return_xcat = list(self.hedged_returns["xcat"].unique())
        else:
            if isinstance(hedged_return_xcat, str):
                hedged_return_xcat = [hedged_return_xcat]
            elif not isinstance(hedged_return_xcat, list):
                raise TypeError("hedged_return_xcat must be a string or a list")
            if not all(isinstance(xcat, str) for xcat in hedged_return_xcat):
                raise TypeError(
                    "All elements in hedged_return_xcat, when a list, must be strings."
                )
            if not (
                set(hedged_return_xcat).issubset(self.hedged_returns["xcat"].unique())
            ):
                raise ValueError(
                    "hedged_return_xcat must be a valid hedged return category within the class instance."
                )

        if cids is None:
            cids = self.hedged_returns["cid"].unique().tolist()
        else:
            if isinstance(cids, str):
                cids = [cids]
            elif not isinstance(cids, list):
                raise TypeError("cids must be a string or a list")
            if not all(isinstance(cid, str) for cid in cids):
                raise TypeError("All elements in cids must be strings.")
            if not all(cid in self.hedged_returns["cid"].unique() for cid in cids):
                raise ValueError(
                    "All cids must be valid cross-section identifiers within the class instance."
                )

        if start is not None and not isinstance(start, str):
            raise TypeError("start must be a string.")

        if end is not None and not isinstance(end, str):
            raise TypeError("end must be a string.")

        if blacklist is not None:
            if not isinstance(blacklist, dict):
                raise TypeError("The blacklist argument must be a dictionary.")
            if len(blacklist) == 0:
                raise ValueError("The blacklist argument must not be empty.")
            if not all([isinstance(key, str) for key in blacklist.keys()]):
                raise TypeError("The keys of the blacklist argument must be strings.")
            if not all(
                [isinstance(value, (list, tuple)) for value in blacklist.values()]
            ):
                raise TypeError("The values of the blacklist argument must be tuples.")
            if not all([len(value) == 2 for value in blacklist.values()]):
                raise ValueError(
                    "The values of the blacklist argument must be tuples of length two."
                )
            if not all(
                [
                    isinstance(date, pd.Timestamp)
                    for value in blacklist.values()
                    for date in value
                ]
            ):
                raise TypeError(
                    "The values of the blacklist argument must be tuples of pandas Timestamps."
                )

        # freqs checks
        if isinstance(freqs, str):
            freqs = [freqs]
        elif not isinstance(freqs, list):
            raise TypeError("freqs must be a string or a list of strings")
        if not all(isinstance(freq, str) for freq in freqs):
            raise TypeError("All elements in freqs must be strings.")
        if not all(freq in ["D", "W", "M", "Q"] for freq in freqs):
            raise ValueError(
                "All elements in freqs must be one of 'D', 'W', 'M' or 'Q'."
            )

        return correlation_types, hedged_return_xcat, cids, freqs

    def get_hedged_returns(
        self,
        hedged_return_xcat = None,
    ):
        """
        Returns a dataframe of out-of-sample hedged returns derived from beta estimation
        processes held within the class instance.

        Parameters
        ----------
        hedged_return_xcat : str or list, optional
            Category name or list of category names
            for the panel of derived hedged returns. If None, information from all
            beta estimation processes held within the class instance is returned.
            Default is None.

        Returns
        -------
        pd.DataFrame
            A dataframe of out-of-sample hedged returns derived from beta estimation
            processes.
        """
        # Checks
        hedged_return_xcat = self._checks_get_hedged_returns(
            hedged_return_xcat=hedged_return_xcat
        )

        if hedged_return_xcat is None:
            hedged_returns = self.hedged_returns
        else:
            hedged_returns = self.hedged_returns[
                self.hedged_returns.xcat.isin(hedged_return_xcat)
            ]

        return QuantamentalDataFrame(
            hedged_returns, _initialized_as_categorical=self.df.InitializedAsCategorical
        ).to_original_dtypes()

    def _checks_get_hedged_returns(
        self,
        hedged_return_xcat,
    ):
        """
        Input checks for the `get_hedged_returns()` method.

        Parameters
        ----------
        hedged_return_xcat : str or list
            Category name or list of category names for the panel of derived hedged
            returns.

        Returns
        -------
        str or list
            Category name or list of category names for the panel of derived hedged
            returns.
        """
        if hedged_return_xcat is not None:
            if isinstance(hedged_return_xcat, str):
                hedged_return_xcat = [hedged_return_xcat]
            elif not isinstance(hedged_return_xcat, list):
                raise TypeError("hedged_return_xcat must be a string or a list")
            if not all(isinstance(xcat, str) for xcat in hedged_return_xcat):
                raise TypeError(
                    "All elements in hedged_return_xcat, when a list, must be strings."
                )
            if not (
                set(hedged_return_xcat).issubset(self.hedged_returns["xcat"].unique())
            ):
                raise ValueError(
                    "hedged_return_xcat must be a valid hedged return category within the class instance."
                )
        return hedged_return_xcat

    def get_betas(
        self,
        beta_xcat = None,
    ):
        """
        Returns a dataframe of estimated betas derived from beta estimation processes
        held within the class instance.

        Parameters
        ----------
        beta_xcat : str or list
            Category name or list of category names for the panel of estimated contract
            betas. If None, information from all beta estimation processes held within
            the class instance is returned. Default is None.

        Returns
        -------
        pd.DataFrame
            A dataframe of estimated betas derived from beta estimation processes.
        """
        # Checks
        beta_xcat = self._checks_get_betas(beta_xcat=beta_xcat)

        if beta_xcat is None:
            betas = self.betas
        else:
            betas = self.betas[self.betas.xcat.isin(beta_xcat)]

        return QuantamentalDataFrame(
            betas, _initialized_as_categorical=self.df.InitializedAsCategorical
        ).to_original_dtypes()

    def _checks_get_betas(
        self,
        beta_xcat,
    ):
        """
        Input checks for the `get_betas()` method.

        Parameters
        ----------
        beta_xcat : str or list
            Category name or list of category names for the panel of estimated contract
            betas.

        Returns
        -------
        str or list
            Category name or list of category names for the panel of estimated contract
            betas.
        """
        if beta_xcat is not None:
            if isinstance(beta_xcat, str):
                beta_xcat = [beta_xcat]
            if not isinstance(beta_xcat, list):
                raise TypeError("beta_xcat must be a string or a list")
            if not all(isinstance(xcat, str) for xcat in beta_xcat):
                raise TypeError(
                    "All elements in beta_xcat, when a list, must be strings."
                )
            if not (set(beta_xcat).issubset(self.betas["xcat"].unique())):
                raise ValueError(
                    "beta_xcat must be a valid beta category within the class instance."
                )
        return beta_xcat

    def _get_mean_abs_corrs(
        self,
        xcat,
        cids,
        df,
        correlation,
    ):
        """
        Calculate mean absolute correlation between a column 'xcat' in a dataframe 'df'
        and the benchmark return (the last column) across all cross-sections in 'cids'.
        The correlation is calculated using the method specified in 'correlation'.

        Parameters
        ----------
        xcat : str
            Category name for the column in the dataframe.
        cids : str
            Cross-sections for which the correlation is calculated.
        df : pd.DataFrame
            Dataframe containing the relevant columns.
        correlation : str
            Type of correlation to calculate.

        Returns
        -------
        float
            Mean absolute correlation between the column 'xcat' and the benchmark return.
        """
        # Get relevant columns
        df_subset = df[[xcat, self.benchmark_xcat]].dropna()

        # Create inner function to calculate the correlation for a given cross-section
        # This is done so that one can groupby cross-section and apply this function directly

        def calculate_correlation(group):
            return abs(group[xcat].corr(group[self.benchmark_xcat], method=correlation))

        # Calculate the mean absolute correlation over all cross sections
        mean_abs_corr = (
            df_subset.groupby("cid", observed=True).apply(calculate_correlation).mean()
        )

        return mean_abs_corr

    def _check_duplicate_results(self, hedged_return_xcat, beta_xcat):
        """
        Check for duplicate results in the class instance and remove them.

        Parameters
        ----------
        hedged_return_xcat : str
            Category name for the panel of out-of-sample hedged returns.
        beta_xcat : str
            Category name for the panel of estimated betas.
        """
        conditions = [
            ("hedged_returns", "xcat", hedged_return_xcat),
            ("betas", "xcat", beta_xcat),
            ("chosen_models", "name", beta_xcat),
        ]

        self._remove_results(conditions)


if __name__ == "__main__":
    from macrosynergy.learning import (
        ExpandingKFoldPanelSplit,
        LinearRegressionSystem,
        neg_mean_abs_corr,
    )
    from macrosynergy.management.simulate import make_qdf

    # Simulate a panel dataset of benchmark and contract returns
    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["BENCH_XR", "CONTRACT_XR"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2015-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2015-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2015-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2015-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["BENCH_XR"] = ["2015-01-01", "2019-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CONTRACT_XR"] = ["2015-01-01", "2019-12-31", 0.1, 1, 0, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    # Initialize the BetaEstimator object
    # Use for the benchmark return: USD_BENCH_XR.
    be = BetaEstimator(
        df=dfd,
        xcats="CONTRACT_XR",
        benchmark_return="USD_BENCH_XR",
        cids=["AUD", "USD"],
    )

    models = {
        "LR": LinearRegressionSystem(min_xs_samples=21 * 1),
    }
    hparam_grid = {"LR": {"fit_intercept": [True, False], "positive": [True, False]}}

    scorer = {"scorer": neg_mean_abs_corr}

    be.estimate_beta(
        beta_xcat="BETA_NSA",
        hedged_return_xcat="HEDGED_RETURN_NSA",
        models=models,
        hyperparameters=hparam_grid,
        scorers=scorer,
        inner_splitters={"expandingkfold": ExpandingKFoldPanelSplit(n_splits=5)},
        search_type="grid",
        cv_summary="median",
        min_cids=1,
        min_periods=21 * 12,
        est_freq="Q",
        n_jobs_outer=1,
        n_jobs_inner=1,
    )

    evaluation_df = be.evaluate_hedged_returns(
        correlation_types=["pearson", "spearman", "kendall"],
        freqs=["W", "M", "Q"],
    )

    be.models_heatmap(name="BETA_NSA")
    # print(evaluation_df)



from .lad_regressors import (
    LADRegressor,
    SignWeightedLADRegressor,
    TimeWeightedLADRegressor,
)

from .ls_regressors import (
    SignWeightedLinearRegression,
    TimeWeightedLinearRegression,
    ModifiedLinearRegression,
    ModifiedSignWeightedLinearRegression,
    ModifiedTimeWeightedLinearRegression,
)

__all__ = [
    "LADRegressor",
    "SignWeightedLADRegressor",
    "TimeWeightedLADRegressor",
    "SignWeightedLinearRegression",
    "TimeWeightedLinearRegression",
    "ModifiedLinearRegression",
    "ModifiedSignWeightedLinearRegression",
    "ModifiedTimeWeightedLinearRegression",
]


import numpy as np
import pandas as pd
import warnings
import datetime
import numbers

from sklearn.base import BaseEstimator, TransformerMixin


class ZnScoreAverager(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        neutral="zero",
        use_signs=False,
    ):
        """
        Point-in-time factor normalization before feature averaging for a conceptual
        parity signal.

        :deprecated: This class is deprecated and will be replaced in a future release
            by a transformer that applies PiT z-scoring to each feature, without
            any averaging.

        Parameters
        ----------
        neutral : str, default='zero'
            Specified neutral value for a normalisation. This can take either 'zero' or
            'mean'. If the neutral level is zero, each feature is standardised by dividing
            by the mean absolute deviation. If the neutral level is mean, each feature is
            normalised by subtracting the mean and dividing by the standard deviation. Any
            statistics are computed according to a point-in-time principle.
        use_signs : bool, default=False
            Boolean to specify whether or not to return the signs of the benchmark signal
            instead of the signal itself.

        Notes
        -----
        When clear priors on underlying market drivers are available, a useful signal
        can be constructed by averaging PiT :math:`z_{n}`-scores of each feature/factor.
        This is often a competitive signal.

        A :math:`z_{n}`-score involves standardising a feature PiT by subtracting a
        `neutral` value and dividing by a measure of historic dispersion.
        In this class, the neutral value can be set to zero or the mean. An option for the
        median will be added in the future. The dispersion measure used is the mean
        absolute deviation for the zero neutral value and the standard deviation for the
        mean neutral value. The dispersion measure is calculated using all training
        information and test information until (and including) that test time, reflecting
        the information available to a portfolio manager at that time.

        We include that test time since features are assumed to lag behind returns, with
        the timestamp in the index representing the date of the returns.
        """

        warnings.warn(
            "ZnScoreAverager is deprecated and will be replaced in a future "
            "release by a transformer that applies PiT z-scoring to each feature, "
            "without any averaging, as per the current implementation.",
            DeprecationWarning,
        )

        if not isinstance(neutral, str):
            raise TypeError("'neutral' must be a string.")

        if neutral not in ["zero", "mean"]:
            raise ValueError("'neutral' must be either 'zero' or 'mean'.")

        if not isinstance(use_signs, bool):
            raise TypeError("'use_signs' must be a boolean.")

        self.neutral = neutral
        self.use_signs = use_signs
        self.n_features_in_ = None
        self.feature_names_in_ = None

    def fit(self, X, y=None):
        """
        Extract relevant standardisation/normalisation statistics.

        Parameters
        ----------
        X : pd.DataFrame
            Input feature matrix.
        y : Any, default=None
            Placeholder for scikit-learn compatibility.
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the ZnScoreAverager must be a pandas "
                "dataframe. If used as part of an sklearn pipeline, ensure that previous "
                "steps return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not X.index.get_level_values(0).dtype == "object":
            raise TypeError("The outer index of X must be strings.")
        if not X.index.get_level_values(1).dtype == "datetime64[ns]":
            raise TypeError("The inner index of X must be datetime.date.")
        if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
            raise ValueError(
                "All columns in the input feature matrix for a panel selector ",
                "must be numeric.",
            )
        if X.isnull().values.any():
            raise ValueError(
                "The input feature matrix for a panel selector must not contain any "
                "missing values."
            )

        # fit

        self.training_n: int = len(X)
        self.n_features_in_ = X.shape[1]
        self.feature_names_in_ = X.columns

        if self.neutral == "mean":
            # calculate the mean and sum of squares of each feature
            means: pd.Series = X.mean(axis=0)
            sum_squares: pd.Series = np.sum(np.square(X), axis=0)
            self.training_means: pd.Series = means
            self.training_sum_squares: pd.Series = sum_squares
        else:
            # calculate the mean absolute deviation of each feature
            # TODO: maybe use the median?
            mads: pd.Series = np.mean(np.abs(X), axis=0)
            self.training_mads: pd.Series = mads

        return self

    def transform(self, X):
        """
        Create an OOS conceptual parity signal by averaging PiT z-scores of features.

        Parameters
        ----------
        X : pd.DataFrame
            Input feature matrix.

        Returns
        -------
        pd.DataFrame
            Output signal.
        """
        # checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the ZnScoreAverager must be a pandas "
                "dataframe. If used as part of an sklearn pipeline, ensure that previous "
                "steps return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")
        if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
            raise ValueError(
                "All columns in the input feature matrix for `ZnScoreAverager` ",
                "must be numeric.",
            )
        if X.isnull().values.any():
            raise ValueError(
                "The input feature matrix for `ZnScoreAverager` must not contain any "
                "missing values."
            )
        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                "The input feature matrix must have the same number of columns as the "
                "training feature matrix."
            )
        if not X.columns.equals(self.feature_names_in_):
            raise ValueError(
                "The input feature matrix must have the same columns as the training "
                "feature matrix."
            )

        X = X.sort_index(level="real_date")

        # transform
        signal_df = pd.DataFrame(index=X.index, columns=["signal"], dtype="float")

        if self.neutral == "zero":
            training_mads: pd.Series = self.training_mads
            training_n: int = self.training_n
            X_abs = X.abs()

            # We obtain a vector with the number of elements in each expanding window.
            # The count does not necessarily increase uniformly, since some cross-sections
            # may be missing for some dates.
            expanding_count = self._get_expanding_count(X)

            # We divide the expanding sum by the number of elements in each expanding window.
            X_test_expanding_mads = (
                X_abs.groupby(level="real_date").sum().expanding().sum()
                / expanding_count
            )
            n_total = training_n + expanding_count

            X_expanding_mads: pd.DataFrame = (
                (training_n) * training_mads + (expanding_count) * X_test_expanding_mads
            ) / n_total

            standardised_X: pd.DataFrame = (X / X_expanding_mads).fillna(0)

        elif self.neutral == "mean":
            training_means: pd.Series = self.training_means
            training_sum_squares: pd.Series = self.training_sum_squares
            training_n: int = self.training_n

            expanding_count = self._get_expanding_count(X)
            n_total = training_n + expanding_count

            X_square = X**2
            X_test_expanding_sum_squares = (
                X_square.groupby(level="real_date").sum().expanding().sum()
            )
            X_test_expanding_mean = (
                X.groupby(level="real_date").sum().expanding().sum() / expanding_count
            )

            X_expanding_means = (
                (training_n) * training_means
                + (expanding_count) * X_test_expanding_mean
            ) / n_total

            X_expanding_sum_squares = (
                training_sum_squares + X_test_expanding_sum_squares
            )
            comp1 = (X_expanding_sum_squares) / (n_total - 1)
            comp2 = 2 * np.square(X_expanding_means) * (n_total) / (n_total - 1)
            comp3 = (n_total) * np.square(X_expanding_means) / (n_total - 1)
            X_expanding_std: pd.Series = np.sqrt(comp1 - comp2 + comp3)
            standardised_X: pd.DataFrame = (
                (X - X_expanding_means) / X_expanding_std
            ).fillna(0)

        else:
            raise ValueError("neutral must be either 'zero' or 'mean'.")

        benchmark_signal: pd.DataFrame = pd.DataFrame(
            np.mean(standardised_X, axis=1), columns=["signal"], dtype="float"
        )
        signal_df.loc[benchmark_signal.index] = benchmark_signal

        if self.use_signs:
            return np.sign(signal_df).astype(int)

        return signal_df

    def _get_expanding_count(self, X):
        """
        Get the number of non-NaN values in each expanding window.

        Parameters
        ----------
        X : pd.DataFrame
            Input feature matrix.

        Returns
        -------
        np.ndarray
            Expanding count.
        """

        return X.groupby(level="real_date").count().expanding().sum().to_numpy()


class PanelPCA(BaseEstimator, TransformerMixin):
    def __init__(self, n_components=None, kaiser_criterion=False, adjust_signs=False):
        """
        PCA transformer for panel data.

        Parameters
        ----------
        n_components : int, float, or None, default=None
            Number of components to keep. If None, all components are kept.
            If an integer, keep the first `n_components` components.
            If a float, keep the top components that explain the percentage of variance
            specified by the float.
        kaiser_criterion : bool, default=False
            If True, only components with eigenvalues greater or equal to one are kept.
        adjust_signs : bool, default=False
            If True, adjust signs of eigenvectors so that projected training features
            are positively correlated with a provided target variable. This is useful for
            consistency when used in a sequential learning pipeline through time.

        Notes
        -----
        (P)rinicpal (C)omponent (A)nalysis is an unsupervised method that projects an
        input feature matrix onto a new basis of orthogonal vectors, called principal
        components. This basis is chosen to be the directions of maximal variability within
        the estimated covariance matrix of the input features. The principal components are
        ordered by the amount of variance they explain. Typically, the first few components
        can be attributed to "signal" whilst the remaining components are "noise".

        Most modern implementations use the singular value decomposition (SVD) to perform
        PCA. We estimate a covariance matrix because it follows that an alternative
        covariance estimator can be used for PCA, which may affect results. However, this
        behaviour will be introduced in a future release.
        """
        if n_components is not None:
            if not isinstance(n_components, numbers.Number) or isinstance(
                n_components, bool
            ):
                raise TypeError("n_components must be a number or None.")
            if n_components <= 0:
                raise ValueError("n_components must be greater than 0.")
            if isinstance(n_components, float) and n_components > 1:
                raise ValueError(
                    "If n_components is a float, it must be between 0 and 1."
                )

        if not isinstance(kaiser_criterion, bool):
            raise TypeError("kaiser_criterion must be a boolean.")

        if not isinstance(adjust_signs, bool):
            raise TypeError("adjust_signs must be a boolean.")

        self.n_components = n_components
        self.kaiser_criterion = kaiser_criterion
        self.adjust_signs = adjust_signs
        self.n_features_in_ = None
        self.feature_names_in_ = None

    def fit(self, X, y=None):
        """
        Fit method to determine an eigenbasis for the PCA.

        Parameters
        ----------
        X : pd.DataFrame
            Input feature matrix.
        y : pd.DataFrame, pd.Series or np.ndarray, default=None
            Target variable.

        Notes
        -----
        The target variable `y` is only ever used to adjust the signs
        of the eigenvectors to ensure consistency of eigenvector signs
        when retrained over time. This does not affect the PCA itself.
        """
        # Checks
        self._check_fit_params(X, y)
        self.n_features_in_ = X.shape[1]
        self.feature_names_in_ = X.columns

        # Estimate covariance matrix and perform eigendecomposition
        covariance_matrix = X.cov()
        evals, evecs = np.linalg.eigh(covariance_matrix.values)

        # Sort eigenvalues and eigenvectors by descending eigenvalue order
        sorted_idx = np.argsort(evals)[::-1]

        # Store eigenvalues and eigenvectors
        self.adjusted_evals = evals[sorted_idx]
        self.adjusted_evecs = evecs[:, sorted_idx]

        if self.kaiser_criterion:
            # Get eigenvalues greater or equal to one
            mask = self.adjusted_evals >= 1
            self.adjusted_evals = self.adjusted_evals[mask]
            self.adjusted_evecs = self.adjusted_evecs[:, mask]
        elif isinstance(self.n_components, int):
            # Keep first n_components components
            self.adjusted_evals = self.adjusted_evals[: self.n_components]
            self.adjusted_evecs = self.adjusted_evecs[:, : self.n_components]
        elif isinstance(self.n_components, float):
            # Keep components that explain a certain percentage of difference.
            variance_explained = self.adjusted_evals / np.sum(self.adjusted_evals)
            cumulative_variance_explained = np.cumsum(variance_explained)
            mask = cumulative_variance_explained <= self.n_components
            self.adjusted_evals = self.adjusted_evals[mask]
            self.adjusted_evecs = self.adjusted_evecs[:, mask]

        # Adjust signs of eigenvectors so that projected data is positively correlated with y
        if self.adjust_signs:
            # Convert y to numpy array
            if isinstance(y, pd.Series):
                y = y.values
            elif isinstance(y, pd.DataFrame):
                y = y.values.reshape(-1)

            # Calculate correlation between projected data and y
            # If correlation is negative, flip the sign of the eigenvector
            for i in range(self.adjusted_evecs.shape[1]):
                if np.corrcoef(X.values @ self.adjusted_evecs[:, i], y)[0, 1] < 0:
                    self.adjusted_evecs[:, i] *= -1

        return self

    def transform(self, X):
        """
        Project input features onto the principal components.

        Parameters
        ----------
        X : pd.DataFrame
            Input feature matrix.

        Returns
        -------
        pd.DataFrame
            Projected features.
        """
        # Checks
        self._check_transform_params(X)

        # Project data onto the principal components
        return pd.DataFrame(
            index=X.index,
            columns=[f"PCA {i+1}" for i in range(self.adjusted_evecs.shape[1])],
            data=X.values @ self.adjusted_evecs,
        )

    def _check_fit_params(self, X, y):
        """
        Checks the input data for the fit method.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        y : pandas.Series or pandas.DataFrame
            The target vector.
        """
        # Checks only necessary on X
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the selector must be a pandas dataframe. ",
                "If used as part of an sklearn pipeline, ensure that previous steps ",
                "return a pandas dataframe.",
            )
        if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
            raise ValueError(
                "All columns in the input feature matrix for PanelPCA",
                " must be numeric.",
            )
        if X.isnull().values.any():
            raise ValueError(
                "The input feature matrix for PanelPCA must not contain any "
                "missing values."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("The input feature matrix for X must be multi-indexed.")
        if not X.index.get_level_values(0).dtype == "object":
            raise TypeError("The outer index of X must be strings.")
        if not X.index.get_level_values(1).dtype == "datetime64[ns]":
            raise TypeError("The inner index of X must be datetime.date.")

        # Checks on y only necessary if adjust_signs is True
        if self.adjust_signs:
            if y is None:
                raise ValueError(
                    "If `adjust_signs` is True, a target variable must be provided. "
                    "PCA is an unsupervised method, so providing a target variable "
                    "does not affect the PCA. Eigenvectors, however, are unique up to "
                    "its sign, meaning that a target variable can be used to ensure "
                    "consistency in the signs of the eigenvectors through time and over repeated runs."
                )
            else:
                if not isinstance(y, (pd.Series, pd.DataFrame, np.ndarray)):
                    raise TypeError(
                        "y must be a pandas Series, DataFrame, or numpy array."
                    )
                if not y.shape[0] == X.shape[0]:
                    raise ValueError("y must have the same number of rows as X.")
                if not isinstance(y, np.ndarray):
                    if not np.issubdtype(y.values.dtype, np.number):
                        raise ValueError("The target vector must be numeric.")
                    if y.isnull().values.any():
                        raise ValueError(
                            "The target vector must not contain any missing values."
                        )
                else:
                    if not np.issubdtype(y.dtype, np.number):
                        raise ValueError("The target vector must be numeric.")
                    if np.isnan(y).any():
                        raise ValueError(
                            "The target vector must not contain any missing values."
                        )

    def _check_transform_params(self, X):
        """
        Input checks for the transform method.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        """
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the PanelPCA must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
            raise ValueError(
                "All columns in the input feature matrix for PanelPCA",
                " must be numeric.",
            )
        if X.isnull().values.any():
            raise ValueError(
                "The input feature matrix for PanelPCA must not contain any "
                "missing values."
            )
        if not X.index.get_level_values(0).dtype == "object":
            raise TypeError("The outer index of X must be strings.")
        if not X.index.get_level_values(1).dtype == "datetime64[ns]":
            raise TypeError("The inner index of X must be datetime.date.")

        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                "The input feature matrix must have the same number of columns as the "
                "training feature matrix."
            )
        if not X.columns.equals(self.feature_names_in_):
            raise ValueError(
                "The input feature matrix must have the same columns as the training "
                "feature matrix."
            )


if __name__ == "__main__":
    import numpy as np
    import pandas as pd

    from macrosynergy.management import (
        categories_df,
        make_qdf,
    )

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", -0.1, 2, 0.8, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {
        "GBP": (
            pd.Timestamp(year=2009, month=1, day=1),
            pd.Timestamp(year=2012, month=6, day=30),
        ),
        "CAD": (
            pd.Timestamp(year=2015, month=1, day=1),
            pd.Timestamp(year=2100, month=1, day=1),
        ),
    }

    train = categories_df(
        df=dfd, xcats=xcats, cids=cids, val="value", blacklist=black, freq="M", lag=1
    ).dropna()
    train = train[
        train.index.get_level_values(1) >= pd.Timestamp(year=2005, month=8, day=1)
    ]

    X_train = train.drop(columns=["XR"])
    y_train = train["XR"]

    # Neutral = zero
    zn = ZnScoreAverager(neutral="zero", use_signs=False)
    zn.fit(X_train, y_train)
    print(zn.transform(X_train))

    # Neutral = mean, use_signs = False
    zn = ZnScoreAverager(neutral="mean", use_signs=False)
    zn.fit(X_train, y_train)
    print(zn.transform(X_train))

    # Neutral = mean, use_signs = True
    zn = ZnScoreAverager(neutral="mean", use_signs=True)
    zn.fit(X_train, y_train)
    print(zn.transform(X_train))